import os
import sys

import quizcomp.args
import quizcomp.converter.qti
import quizcomp.quiz
import quizcomp.util.cli

def run(args):
    if (not os.path.exists(args.path)):
        raise ValueError(f"Provided path '{args.path}' does not exist.")

    if (not os.path.isfile(args.path)):
        raise ValueError(f"Provided path '{args.path}' is not a file.")

    quiz = quizcomp.quiz.Quiz.from_path(args.path)

    out_path = quizcomp.util.cli.resolve_out_arg(args.out, f'{quiz.title}.qti.zip')

    converter = quizcomp.converter.qti.QTITemplateConverter(canvas = args.canvas)
    converter.convert_quiz(quiz, out_path = out_path)

    return 0

def _get_parser():
    parser = quizcomp.args.Parser(description =
        "Parse a quiz and upload the quiz to Canvas.")

    parser.add_argument('path', metavar = 'PATH',
        type = str,
        help = 'The path to a quiz JSON file.')

    parser.add_argument('--canvas', dest = 'canvas',
        action = 'store_true', default = False,
        help = 'Create the QTI with Canvas-specific tweaks (default: %(default)s).')

    quizcomp.util.cli.add_out_arg(parser, '<title>.qti.zip')

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import logging

DEFAULT_LOGGING_LEVEL = logging.getLevelName(logging.INFO)
DEFAULT_LOGGING_FORMAT = '%(asctime)s [%(levelname)-8s] - %(filename)s:%(lineno)s -- %(message)s'

LEVELS = [
    logging.getLevelName(logging.DEBUG),
    logging.getLevelName(logging.INFO),
    logging.getLevelName(logging.WARNING),
    logging.getLevelName(logging.ERROR),
    logging.getLevelName(logging.CRITICAL),
]

def init(level = DEFAULT_LOGGING_LEVEL, format = DEFAULT_LOGGING_FORMAT, **kwargs):
    """
    Initialize or re-initialize the logging infrastructure.
    """

    logging.basicConfig(level = level, format = format, force = True)

    # Ignore logging from third-party libraries.
    logging.getLogger("git").setLevel(logging.WARNING)
    logging.getLogger("markdown_it").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

def set_cli_args(parser):
    """
    Set common logging args in an argparse parser.
    """

    parser.add_argument('--log-level', dest = 'log_level',
        action = 'store', type = str, default = logging.getLevelName(logging.INFO),
        choices = LEVELS,
        help = 'The logging level (default: %(default)s).')

    return parser

def init_from_args(args):
    """
    Take in args from a parser that was passed to set_cli_args(),
    and call init() with the appropriate arguments.
    """

    init(args.log_level)

    return args

# Load the default logging when this module is loaded.
init()



"""
Convert a quiz into QTI using templates.
"""

import html
import logging
import os
import pathlib
import shutil
import warnings

import bs4

import quizcomp.constants
import quizcomp.converter.template
import quizcomp.util.dirent

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DEFAULT_TEMPLATE_DIR = os.path.join(THIS_DIR, '..', 'data', 'templates', 'edq-qti')

QUESTION_TYPE_MAP = {
    # Direct Mappings
    quizcomp.constants.QUESTION_TYPE_ESSAY: 'essay_question',
    quizcomp.constants.QUESTION_TYPE_FIMB: 'fill_in_multiple_blanks_question',
    quizcomp.constants.QUESTION_TYPE_MATCHING: 'matching_question',
    quizcomp.constants.QUESTION_TYPE_MA: 'multiple_answers_question',
    quizcomp.constants.QUESTION_TYPE_MCQ: 'multiple_choice_question',
    quizcomp.constants.QUESTION_TYPE_MDD: 'multiple_dropdowns_question',
    quizcomp.constants.QUESTION_TYPE_NUMERICAL: 'numerical_question',
    quizcomp.constants.QUESTION_TYPE_TEXT_ONLY: 'text_only_question',
    quizcomp.constants.QUESTION_TYPE_TF: 'true_false_question',
    # Indirect Mappings
    quizcomp.constants.QUESTION_TYPE_FITB: 'short_answer_question',
    quizcomp.constants.QUESTION_TYPE_SA: 'essay_question',
}

TEMPLATE_FILENAME_ASSESSMENT_META = 'qti_assessment_meta.template'
TEMPLATE_FILENAME_MANIFEST = 'qti_imsmanifest.template'

OUT_DIR_IMAGES = 'images'
OUT_DIR_QUIZ = 'quiz'
OUT_FILENAME_QUIZ = 'quiz.xml'
OUT_FILENAME_ASSESSMENT_META = 'assessment_meta.xml'
OUT_FILENAME_MANIFEST = 'imsmanifest.xml'

DEFAULT_ID_DELIM = '.'
CANVAS_ID_DELIM = 'f'

class QTITemplateConverter(quizcomp.converter.template.TemplateConverter):
    def __init__(self, template_dir = DEFAULT_TEMPLATE_DIR, canvas = False, **kwargs):
        parser_format_options = {}
        id_delim = DEFAULT_ID_DELIM

        if (canvas):
            parser_format_options = {
                'image_path_callback': self._store_images,
                'force_raw_image_src': True,
            }
            id_delim = CANVAS_ID_DELIM

        super().__init__(quizcomp.constants.FORMAT_HTML, template_dir,
                parser_format_options = parser_format_options,
                id_delim = id_delim,
                jinja_filters = {
                    'to_xml': _to_xml,
                },
                jinja_globals = {
                    'canvas': canvas,
                },
                **kwargs)

        self.canvas = canvas

    def convert_variant(self, variant, **kwargs):
        # Parse and format the XML.
        text = super(QTITemplateConverter, self).convert_variant(variant, **kwargs)
        return self._format_xml(text)

    def modify_question_context(self, context, question, variant):
        context['question']['mapped_question_type'] = QUESTION_TYPE_MAP[question.question_type]
        return context

    def convert_quiz(self, quiz, out_path = None, **kwargs):
        if (out_path is None):
            out_path = f'{quiz.title}.qti.zip'

        temp_base_dir = quizcomp.util.dirent.get_temp_path(prefix = 'quizcomp-qti-')

        temp_out_path = os.path.join(temp_base_dir, 'out.zip')
        temp_dir = os.path.join(temp_base_dir, quiz.title)

        quiz_dir = os.path.join(temp_dir, OUT_DIR_QUIZ)
        os.makedirs(quiz_dir, exist_ok = True)

        if (self.canvas):
            self.image_base_dir = os.path.join(temp_dir, OUT_DIR_IMAGES)
            os.makedirs(self.image_base_dir)

        path = os.path.join(quiz_dir, OUT_FILENAME_QUIZ)
        text = super(QTITemplateConverter, self).convert_quiz(quiz, **kwargs)
        quizcomp.util.dirent.write_file(path, self._format_xml(text))

        self._convert_assessment_meta(quiz, quiz_dir)
        self._convert_manifest(quiz, temp_dir)

        self._create_zip(quiz, temp_out_path, out_path, temp_dir)

        logging.info("Created QTI quiz at '%s'." % (out_path))
        return path

    def _create_zip(self, quiz, temp_out_path, out_path, temp_dir):
        shutil.make_archive(os.path.splitext(temp_out_path)[0], 'zip', os.path.dirname(temp_dir), os.path.basename(temp_dir))
        quizcomp.util.dirent.copy_dirent(temp_out_path, out_path)

    def _format_xml(self, text):
        warnings.filterwarnings('ignore', category = bs4.builder.XMLParsedAsHTMLWarning)
        document = bs4.BeautifulSoup(text, 'html.parser')
        return document.prettify(formatter = bs4.formatter.HTMLFormatter(indent = 4))

    def _convert_assessment_meta(self, quiz, out_dir):
        template = self.env.get_template(TEMPLATE_FILENAME_ASSESSMENT_META)

        quiz_context = quiz.to_dict()

        description_text = quiz.description.document.to_format(self.format)
        description_text = "<p>%s</p><br /><hr /><p>Version: %s</p>" % (description_text, quiz.version)

        if (self.canvas):
            description_text = html.escape(description_text)

        quiz_context['description_text'] = description_text

        text = template.render(quiz = quiz_context)

        if (not self.canvas):
            # Canvas has some very strange and undocumented formatting requirements for the assessment meta file.
            # Do not format/pretty when working with Canvas.
            text = self._format_xml(text)

        path = os.path.join(out_dir, OUT_FILENAME_ASSESSMENT_META)
        quizcomp.util.dirent.write_file(path, text)

    def _convert_manifest(self, quiz, out_dir):
        template = self.env.get_template(TEMPLATE_FILENAME_MANIFEST)

        data = {
            'quiz': quiz.to_dict(),
            'files': [],
        }

        for (old_path, new_path) in self.image_paths.items():
            data['files'].append({
                'type': 'image',
                'id': os.path.splitext(os.path.basename(new_path))[0],
                'raw_path': new_path,
                'path': '/'.join([quiz.title, OUT_DIR_IMAGES, os.path.basename(new_path)]),
                'filename': os.path.basename(new_path),
            })

        text = template.render(**data)
        text = self._format_xml(text)

        path = os.path.join(out_dir, OUT_FILENAME_MANIFEST)
        quizcomp.util.dirent.write_file(path, text)

    def _store_images(self, link, base_dir):
        """
        Override the final path that is returned to instead point to the Canvas path.
        Note that this method should only be called when (self.canvas == True).
        """

        path = super(QTITemplateConverter, self)._store_images(link, base_dir)

        quiz_name = os.path.basename(os.path.dirname(self.image_base_dir))
        filename = os.path.basename(path)

        return '/'.join(['$IMS-CC-FILEBASE$', quiz_name, OUT_DIR_IMAGES, filename])

def _to_xml(item):
    """
    Convert the item to an XML string.
    """

    if (item is None):
        return ''

    if (isinstance(item, bool)):
        return str(item).lower()

    return str(item)



import quizcomp.common
import quizcomp.parser.public
import quizcomp.util.serial

class ParsedTextWithFeedback(quizcomp.parser.public.ParsedText):
    def __init__(self, parsed_text, feedback = None):
        super().__init__(parsed_text.text, parsed_text.document)

        if ((feedback is not None) and (not isinstance(feedback, quizcomp.parser.public.ParsedText))):
            raise quizcomp.common.QuizValidationError("Text feedback must be quizcomp.parser.public.ParsedText, found '%s'." % (str(type(feedback))))

        self.feedback = feedback

    def has_feedback(self):
        return (self.feedback is not None)

    def to_pod(self, skip_feedback = False, force_dict = False, **kwargs):
        if (skip_feedback or (self.feedback is None)):
            if (force_dict):
                return {'text': self.text}
            else:
                return self.text

        return {
            'text': self.text,
            'feedback': self.feedback.to_pod(),
        }

class ParsedTextChoice(ParsedTextWithFeedback):
    """
    A multiple answer/choice option.
    """

    def __init__(self, parsed_text_with_feedback, correct):
        super().__init__(parsed_text_with_feedback, feedback = parsed_text_with_feedback.feedback)

        if (not isinstance(correct, bool)):
            raise quizcomp.common.QuizValidationError("Choice 'correct' field must be boolean, found '%s'." % (str(type(correct))))

        self.correct = correct

    def is_correct(self):
        return self.correct

    def to_pod(self, **kwargs):
        value = super().to_pod(force_dict = True, **kwargs)
        value['correct'] = self.correct
        return value

class NumericChoice(quizcomp.util.serial.PODSerializer):
    """
    Numeric choices have no parsed text (aside from optional feedback).
    """

    def __init__(self, type, margin = None, min = None, max = None, value = None, precision = None, feedback = None):
        self.type = type
        self.margin = margin
        self.min = min
        self.max = max
        self.value = value
        self.precision = precision

        if ((feedback is not None) and (not isinstance(feedback, quizcomp.parser.public.ParsedText))):
            raise quizcomp.common.QuizValidationError("Text feedback must be quizcomp.parser.public.ParsedText, found '%s'." % (str(type(feedback))))

        self.feedback = feedback

    def has_feedback(self):
        return (self.feedback is not None)

    def to_pod(self, skip_feedback = False, **kwargs):
        data = self.__dict__.copy()

        for (key, value) in list(data.items()):
            if (value is None):
                del data[key]

        if (skip_feedback and ('feedback' in data)):
            del data['feedback']

        if ('feedback' in data):
            data['feedback'] = data['feedback'].to_pod()

        return data



import os

import quizcomp.common
import quizcomp.question.base
import quizcomp.uploader.canvas
import quizcomp.util.json
import tests.base

CANVAS_FILENAME = 'canvas.json'

CANVAS_TEST_GROUP_ID = 0
CANVAS_TEST_INDEX = 0

class QuestionsTest(tests.base.BaseTest):
    """
    Test parsing/generating all questions in the 'tests/questions/good' directory.
    A 'question.json' indicates a question that should be parsed.
    A 'canvas.json' in the same directory indicates that the question
    should also be checked for it's Canvas format.

    Test that questions in 'tests/questions/bad' do not parse.
    """

    pass

def _add_question_tests():
    good_paths, bad_paths = tests.base.discover_question_tests()

    for path in good_paths:
        try:
            _add_question_test(path)
        except Exception as ex:
            raise ValueError("Failed to parse test case '%s'." % (path)) from ex

    for path in bad_paths:
        try:
            _add_fail_question_test(path)
        except Exception as ex:
            raise ValueError("Failed to parse failing test case '%s'." % (path)) from ex

def _add_question_test(path):
    base_test_name = os.path.splitext(os.path.basename(os.path.dirname(path)))[0]

    test_name = 'test_question_parse_' + base_test_name
    setattr(QuestionsTest, test_name, _get_question_parse_test_method(path))

    test_name = 'test_question_reparse_' + base_test_name
    setattr(QuestionsTest, test_name, _get_question_reparse_test_method(path))

    canvas_path = os.path.join(os.path.dirname(path), CANVAS_FILENAME)
    if (os.path.exists(canvas_path)):
        test_name = 'test_question_canvas_' + base_test_name
        setattr(QuestionsTest, test_name, _get_question_canvas_test_method(path, canvas_path))

def _get_question_parse_test_method(path):
    """
    Get a test for just parsing a question file.
    """

    def __method(self):
        question = quizcomp.question.base.Question.from_path(path)
        self.assertIsNotNone(question)

    return __method

def _get_question_reparse_test_method(path):
    """
    Get a test for parsing a question file, converting the question to a dict, then re-parsing the same question.
    """

    def __method(self):
        question = quizcomp.question.base.Question.from_path(path)
        question_data = question.to_dict()

        new_question = quizcomp.question.base.Question.from_dict(question_data)
        new_question_data = new_question.to_dict()

        self.assertJSONDictEqual(question_data, new_question_data)

    return __method

def _get_question_canvas_test_method(path, canvas_path):
    def __method(self):
        question = quizcomp.question.base.Question.from_path(path)
        canvas_info = quizcomp.uploader.canvas._create_question_json(CANVAS_TEST_GROUP_ID, question, CANVAS_TEST_INDEX)

        with open(canvas_path, 'r') as file:
            expected_canvas_info = quizcomp.util.json.load(file)

        self.assertJSONDictEqual(expected_canvas_info, canvas_info)

    return __method

def _add_fail_question_test(path):
    base_test_name = os.path.splitext(os.path.basename(os.path.dirname(path)))[0]

    test_name = 'test_question_fail_' + base_test_name
    setattr(QuestionsTest, test_name, _get_question_fail_test_method(path))

def _get_question_fail_test_method(path):
    def __method(self):
        with self.assertRaises(quizcomp.common.QuizValidationError):
            quizcomp.question.base.Question.from_path(path)

    return __method

_add_question_tests()



import quizcomp.common
import quizcomp.constants
import quizcomp.question.base

class Matching(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_MATCHING):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        self._check_type(self.answers, dict, "'answers' key")

        self._validate_matches()
        self._validate_distractors()

    def _validate_matches(self):
        if ('matches' not in self.answers):
            raise quizcomp.common.QuestionValidationError(self, "Matching 'answers' value is missing the 'matches' field.")

        matches = self.answers['matches']
        new_matches = []

        for i in range(len(matches)):
            match = matches[i]

            if (isinstance(match, list)):
                if (len(match) != 2):
                    raise quizcomp.common.QuestionValidationError(self, f"Expected exactly two items for a match list, found {len(match)} items at element {i}.")

                match = {
                    'left': match[0],
                    'right': match[1],
                }

            keys = ['left', 'right']
            for key in keys:
                if (key not in match):
                    raise quizcomp.common.QuestionValidationError(self, "Missing key '{key}' for for match item {i}.")

            new_matches.append({
                'left': self._validate_text_item(match['left'], "Left value for match item %d" % (i)),
                'right': self._validate_text_item(match['right'], "Right value for match item %d" % (i)),
            })

        self.answers['matches'] = new_matches

    def _validate_distractors(self):
        if ('distractors' not in self.answers):
            self.answers['distractors'] = []

        distractors = self.answers['distractors']
        new_distractors = []

        for i in range(len(distractors)):
            new_distractors.append(self._validate_text_item(distractors[i], "distractor at index %d" % (i),
                    clean_whitespace = True))

        self.answers['distractors'] = new_distractors

    def _shuffle(self, rng):
        # Shuffling matching is special because it requires additional shuffling support at the converter level.
        self.answers['shuffle'] = True
        self.answers['shuffle_seed'] = rng.randint(0, 2 ** 64)



import quizcomp.common
import quizcomp.constants
import quizcomp.question.base
import quizcomp.question.common

class Numerical(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_NUMERICAL):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        self._check_type(self.answers, list, "'answers' key")

        for i in range(len(self.answers)):
            answer = self.answers[i]
            label = "'answers' value index %d" % (i)

            self._check_type(answer, dict, label)

            if ('type' not in answer):
                raise quizcomp.common.QuestionValidationError(self, f"Missing key ('type') for {label}.")

            if (answer['type'] == quizcomp.constants.NUMERICAL_ANSWER_TYPE_EXACT):
                required_keys = ['value']

                if ('margin' not in answer):
                    answer['margin'] = 0
            elif (answer['type'] == quizcomp.constants.NUMERICAL_ANSWER_TYPE_RANGE):
                required_keys = ['min', 'max']
            elif (answer['type'] == quizcomp.constants.NUMERICAL_ANSWER_TYPE_PRECISION):
                required_keys = ['value', 'precision']
            else:
                raise quizcomp.common.QuestionValidationError(self, f"Unknown numerical answer type: '{answer['type']}'.")

            for key in required_keys:
                if (key not in answer):
                    raise quizcomp.common.QuestionValidationError(self, f"Missing required key '{key}' for numerical answer type '{answer['type']}'.")

            feedback = self._validate_feedback_item(answer.get('feedback', None), label)
            if (feedback is not None):
                answer['feedback'] = feedback
            else:
                answer.pop('feedback', None)

            self.answers[i] = quizcomp.question.common.NumericChoice(**answer)



import quizcomp.common
import quizcomp.constants
import quizcomp.question.base

class FITB(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_FITB):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        """
        Set up the answers to look like fill in multiple blanks with an empty key.
        """

        # Standardize an alternate format.
        # Note that this is the format we will eventually output,
        # but we want to validate the user-facing format so errors are more clear.
        if (isinstance(self.answers, dict)):
            if (len(self.answers) != 1):
                raise quizcomp.common.QuestionValidationError(self, "Dict format should have exactly one entry (''), found %d entries." % (len(self.answers)))

            if ('' not in self.answers):
                raise quizcomp.common.QuestionValidationError(self, "Dict format does not have required key '' (empty string).")

            self.answers = self.answers['']

            # If this was already in the full FIMB format, then we need to pull out the values.
            if ((isinstance(self.answers, dict)) and ('values' in self.answers)):
                self.answers = self.answers['values']

        self._check_type(self.answers, list, f"'answers' value")

        if (len(self.answers) == 0):
            raise quizcomp.common.QuestionValidationError(self, "Expected 'answers' value to be non-empty.")

        self.answers = {'': self.answers}

        self._validate_fimb_answers()



import abc

import markdown_it.renderer

import quizcomp.parser.ast
import quizcomp.parser.common
import quizcomp.parser.style

class QuizComposerRendererBase(markdown_it.renderer.RendererProtocol, abc.ABC):
    def render(self, tokens, options, env):
        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})

        # Work with an AST instead of tokens.
        ast = quizcomp.parser.ast.build(tokens)

        return self._render_node(ast, context)

    def clean_final(self, text, context):
        """
        Last chance for cleaning before leaving the renderer.
        """

        return text.strip()

    def _render_node(self, node, context):
        """
        Route rendering to a the method '_render_<type>(self, node, context)', e.g.: '_image'.
        """

        method_name = '_' + node.type()
        method = getattr(self, method_name, None)
        if (method is None):
            raise TypeError("Could not find TeX render method: '%s'." % (method_name))

        return method(node, context)

    def _root(self, node, context):
        content = "\n\n".join([self._render_node(child, context) for child in node.children()])
        content = self.clean_final(content, context)
        return content

    def _container_block(self, node, context):
        # Pull any style attatched to this block and put it in a copy of the context.
        context, _, _ = quizcomp.parser.common.handle_block_style(node, context)
        return "\n\n".join([self._render_node(child, context) for child in node.children()])

    def _paragraph(self, node, context):
        return "\n".join([self._render_node(child, context) for child in node.children()])

    def _inline(self, node, context):
        return ''.join([self._render_node(child, context) for child in node.children()])

    @abc.abstractmethod
    def _text(self, node, context):
        pass

    @abc.abstractmethod
    def _softbreak(self, node, context):
        pass

    @abc.abstractmethod
    def _hardbreak(self, node, context):
        pass

    @abc.abstractmethod
    def _em(self, node, context):
        pass

    @abc.abstractmethod
    def _strong(self, node, context):
        pass

    @abc.abstractmethod
    def _fence(self, node, context):
        pass

    @abc.abstractmethod
    def _code_block(self, node, context):
        """
        This token is poorly named, it is actually an indented code block.
        Treat it like a fence with no info string.
        """

        pass

    @abc.abstractmethod
    def _code_inline(self, node, context):
        pass

    @abc.abstractmethod
    def _math_block(self, node, context):
        pass

    @abc.abstractmethod
    def _math_inline(self, node, context):
        pass

    @abc.abstractmethod
    def _image(self, node, context):
        pass

    @abc.abstractmethod
    def _link(self, node, context):
        pass

    @abc.abstractmethod
    def _placeholder(self, node, context):
        pass

    @abc.abstractmethod
    def _table(self, node, context):
        pass

    @abc.abstractmethod
    def _thead(self, node, context):
        pass

    @abc.abstractmethod
    def _tbody(self, node, context):
        pass

    @abc.abstractmethod
    def _tr(self, node, context):
        pass

    @abc.abstractmethod
    def _th(self, node, context):
        pass

    @abc.abstractmethod
    def _td(self, node, context):
        pass

    @abc.abstractmethod
    def _bullet_list(self, node, context):
        pass

    @abc.abstractmethod
    def _ordered_list(self, node, context):
        pass

    @abc.abstractmethod
    def _list_item(self, node, context):
        pass

    @abc.abstractmethod
    def _hr(self, node, context):
        pass

    @abc.abstractmethod
    def _heading(self, node, context):
        pass

    @abc.abstractmethod
    def _blockquote(self, node, context):
        pass

    def parse_heading_level(self, node):
        # Parse the level out of the HTML tag.
        tag = node.get('tag', None)
        if (tag is None):
            raise ValueError("Failed to find a heading's level.")

        try:
            level = int(tag[1])
        except Exception as ex:
            raise ValueError("Failed to parse heading level from '%s'." % (tag)) from ex

        return level

def get_renderer(options):
    return QuizComposerRendererTex(), options



import base64

DEFAULT_ENCODING = 'utf-8'

def to_base64(data, encoding = DEFAULT_ENCODING):
    if (isinstance(data, str)):
        data = data.encode(encoding)

    data = base64.standard_b64encode(data)
    return data.decode(encoding)

def from_base64(data, encoding = DEFAULT_ENCODING):
    return base64.b64decode(data.encode(encoding), validate = True)



class LayeredDict(object):
    """
    A Layered dict that holds sets of values as layers.
    Containers can push_layer() to add a new layer,
    and pop_layer() to remove the most recent layer.
    When a key is fetched, the layers will be checked (recent to oldest)
    and the first found instance will be returned (allowing new layers to mask old values).
    The initial layer cannot be popped (an error will be raised).
    """

    def __init__(self):
        self._layers = [{}]

    def push_layer(self):
        return self._layers.append({})

    def pop_layer(self):
        if (len(self._layers) == 1):
            raise ValueError("Cannot pop initial layer.")

        return self._layers.pop()

    def get(self, key, default_value = None):
        for layer in reversed(self._layers):
            if (key in layer):
                return layer[key]

        return default_value

    def set(self, key, item, base_layer = False):
        layer = self._layers[-1]
        if (base_layer):
            layer = self._layers[0]

        layer[key] = item

    def delete(self, key, base_layer = False):
        if (base_layer):
            del self._layers[-1][key]
            return

        for layer in reversed(self._layers):
            if (key in layer):
                del layer[key]

    def items(self):
        items = {}

        for layer in self._layers:
            for (key, value) in layer.items():
                items[key] = value

        return items.items()

    def to_pod(self):
        return self._layers

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, item):
        return self.set(key, item)

    def __delitem__(self, key):
        return self.delete(key)

    def __contains__(self, key):
        for layer in reversed(self._layers):
            if (key in layer):
                return True

        return False



import quizcomp.util.encoding

def to_base64(path):
    with open(path, 'rb') as file:
        data = file.read()

    return quizcomp.util.encoding.to_base64(data)

def from_base64(data, path):
    data = quizcomp.util.encoding.from_base64(data)
    with open(path, 'wb') as file:
        file.write(data)




"""
The `quizcomp.cli` package contains tools for running the quizcomp.
Each package can be invoked to list the tools (or subpackages) it contains.
Each tool includes a help prompt that accessed with the `-h`/`--help` flag.
"""

import sys

import quizcomp.util.cli

def main():
    return quizcomp.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())



import datetime
import logging
import os
import random
import string
import traceback

import quizcomp.converter.tex
import quizcomp.latex
import quizcomp.util.dirent
import quizcomp.util.json
import quizcomp.quiz

OPTIONS_FILENAME = 'options.json'

def make_with_args(args, **kwargs):
    """
    Use a standard args object from set_cli_args() to make a PDF quiz.
    """

    if (not os.path.exists(args.path)):
        raise ValueError(f"Provided path '{args.path}' does not exist.")

    if (not os.path.isfile(args.path)):
        raise ValueError(f"Provided path '{args.path}' is not a file.")

    if ((args.variants < 1) or (args.variants >= len(string.ascii_uppercase))):
        raise ValueError("Number of variants must be in [1, %d), found %d." % (len(string.ascii_uppercase), args.variants))

    return make_with_path(args.path, base_out_dir = args.out_dir, seed = args.seed, num_variants = args.variants,
            skip_key = args.skip_key, skip_tex = args.skip_tex, skip_pdf = args.skip_pdf,
            **kwargs)

def make_with_path(quiz_path, **kwargs):
    quiz = quizcomp.quiz.Quiz.from_path(quiz_path)
    return make(quiz, quiz_path = quiz_path, **kwargs)

def make(quiz,
        quiz_path = None, base_out_dir = None,
        seed = None, num_variants = 1, write_options = True,
        skip_key = False, skip_tex = False, skip_pdf = False,
        **kwargs):
    if (base_out_dir is None):
        base_out_dir = quizcomp.util.dirent.get_temp_path(prefix = 'quizcomp_pdf_', rm = False)

    out_dir = os.path.join(base_out_dir, quiz.title)
    os.makedirs(out_dir, exist_ok = True)

    logging.info("Writing TeX/PDF quiz ('%s') to '%s'.", quiz.title, out_dir)

    if (seed is None):
        seed = random.randint(0, 2**64)

    options = {
        'create_time': datetime.datetime.now().isoformat(),
        'seed': seed,
        'out_dir': out_dir,
        'quiz': {
            'path': quiz_path,
            'title': quiz.title,
            'version': quiz.version,
        },
        'variants': [],
    }

    logging.info("Using seed %d.", seed)
    rng = random.Random(seed)

    variants = []
    for i in range(num_variants):
        variant_id = None
        if (num_variants > 1):
            variant_id = string.ascii_uppercase[i]

        variant_seed = rng.randint(0, 2**64)
        variant = quiz.create_variant(identifier = variant_id, seed = variant_seed)
        variants.append(variant)

        out_path = os.path.join(out_dir, "%s.json" % (variant.title))
        quizcomp.util.dirent.write_file(out_path, variant.to_json())

        make_pdf(variant, out_dir = out_dir, is_key = False, skip_tex = skip_tex, skip_pdf = skip_pdf)

        title = variant.title

        # Always create an answer key.
        has_key = False
        if (not skip_key):
            try:
                variant.title = "%s -- Answer Key" % (title)
                make_pdf(variant, out_dir = out_dir, is_key = True, skip_tex = skip_tex, skip_pdf = skip_pdf)
                has_key = True
            except Exception as ex:
                logging.warning("Failed to generate answer key for '%s'.", title)
                logging.debug(traceback.format_exc())
            finally:
                variant.title = title

        options['variants'].append({
            'id': variant_id,
            'title': title,
            'seed': variant_seed,
            'has_key': has_key,
        })

        logging.info("Completed variant: '%s'.", title)

    if (write_options):
        path = os.path.join(out_dir, OPTIONS_FILENAME)
        with open(path, 'w') as file:
            quizcomp.util.json.dump(options, file, indent = 4)

    return (quiz, variants, options)

def make_pdf(variant,
        out_dir = None, is_key = False,
        skip_tex = False, skip_pdf = False):
    if (out_dir is None):
        out_dir = quizcomp.util.dirent.get_temp_path(prefix = 'quizcomp_pdf_', rm = False)

    image_relative_root = os.path.join('images', variant.title)
    image_dir = os.path.join(out_dir, image_relative_root)

    out_path = os.path.join(out_dir, "%s.tex" % (variant.title))

    if (not skip_tex):
        converter = quizcomp.converter.tex.TexTemplateConverter(answer_key = is_key,
                image_base_dir = image_dir, image_relative_root = image_relative_root, cleanup_images = True)
        content = converter.convert_variant(variant)

        quizcomp.util.dirent.write_file(out_path, content)

    if (not skip_pdf):
        quizcomp.latex.compile(out_path)

    return out_dir

def set_cli_args(parser):
    parser.add_argument('path', metavar = 'PATH',
        type = str,
        help = 'The path to a quiz json file.')

    parser.add_argument('--variants', dest = 'variants',
        action = 'store', type = int, default = 1,
        help = 'The number of quiz variants to create (default: %(default)s).')

    parser.add_argument('--outdir', dest = 'out_dir',
        action = 'store', type = str, default = '.',
        help = 'The directory to put the quiz creation output (which will be another directory) (default: %(default)s).')

    parser.add_argument('--skip-key', dest = 'skip_key',
        action = 'store_true', default = False,
        help = 'Skip creating the answer key (default: %(default)s).')

    parser.add_argument('--skip-tex', dest = 'skip_tex',
        action = 'store_true', default = False,
        help = 'Skip creating TeX files (assumes the TeX files already exist) (default: %(default)s).')

    parser.add_argument('--skip-pdf', dest = 'skip_pdf',
        action = 'store_true', default = False,
        help = 'Skip compiling PDFs from TeX (assumes the PDFs already exist) (default: %(default)s).')

    parser.add_argument('--seed', dest = 'seed',
        action = 'store', type = int, default = None,
        help = 'The random seed to use (defaults to a random seed).')

    return parser



import os
import shutil
import urllib.parse

import requests

def get(url, headers = {}, options = {}):
    response = requests.get(url, headers = headers, **options)
    response.raise_for_status()
    return response

def get_file(url, out_path, headers = {}):
    if (os.path.isdir(out_path)):
        parts = urllib.parse.urlsplit(url)
        filename = os.path.basename(parts.path)
        out_path = os.path.join(out_path, filename)

    response = get(url, headers = headers, options = {'stream': True})
    with open(out_path, 'wb') as file:
        shutil.copyfileobj(response.raw, file)

    return out_path



import logging

import quizcomp.log
import tests.base

class TestLog(tests.base.BaseTest):
    def test_base_levels(self):
        quizcomp.log.init()

        print()
        logging.debug("log test: debug")
        logging.info("log test: info")
        logging.warning("log test: warning")
        logging.error("log test: error")



import os

import quizcomp.converter.convert
import tests.base

class ConverterQuestionsTest(tests.base.BaseTest):
    """
    Use each standard converter/format on each "good" quiz question.
    """

    # Cache questions (by path) that have already been parased.
    _question_cache = {}

    def _get_question(self, path):
        path = os.path.abspath(path)
        if (path in ConverterQuestionsTest._question_cache):
            return ConverterQuestionsTest._question_cache[path]

        question = quizcomp.question.base.Question.from_path(path)
        ConverterQuestionsTest._question_cache[path] = question

        return question

def _add_converter_tests():
    good_paths, _ = tests.base.discover_question_tests()

    for path in good_paths:
        base_test_name = os.path.splitext(os.path.basename(os.path.dirname(path)))[0]

        for format_name in quizcomp.converter.convert.SUPPORTED_FORMATS:
            for key in [True, False]:
                    test_name = 'test_converter_question__%s__%s__key_%s' % (base_test_name, format_name, str(key))
                    setattr(ConverterQuestionsTest, test_name, _get_template_test(path, format_name, key))

def _get_template_test(path, format_name, key):
    def __method(self):
        constructor_args = {'answer_key': key}

        question = self._get_question(path)
        content = quizcomp.converter.convert.convert_question(question, format = format_name,
                constructor_args = constructor_args)

        self.assertTrue(len(content) > 10)

    return __method

_add_converter_tests()



import copy
import os
import re
import types

import quizcomp.constants
import quizcomp.parser.ast
import quizcomp.parser.render
import quizcomp.parser.common
import quizcomp.util.json

class ParsedDocument(object):
    def __init__(self, tokens, base_dir = '.'):
        self._tokens = tokens
        self._context = {
            quizcomp.parser.common.BASE_DIR_KEY: base_dir,
        }

    def set_context_value(self, key, value):
        self._context[key] = value

    def to_canvas(self, **kwargs):
        return self._render(quizcomp.constants.FORMAT_CANVAS, **kwargs)

    def to_md(self, **kwargs):
        return self._render(quizcomp.constants.FORMAT_MD, **kwargs)

    def to_tex(self, **kwargs):
        return self._render(quizcomp.constants.FORMAT_TEX, **kwargs)

    def to_text(self, **kwargs):
        return self._render(quizcomp.constants.FORMAT_TEXT, **kwargs)

    def to_html(self, **kwargs):
        return self._render(quizcomp.constants.FORMAT_HTML, **kwargs)

    def _render(self, format, **kwargs):
        context = quizcomp.parser.common.prep_context(self._context, options = kwargs)
        env = {quizcomp.parser.common.CONTEXT_ENV_KEY: context}
        return quizcomp.parser.render.render(format, self._tokens, env = env, **kwargs)

    def to_pod(self, include_metadata = True, **kwargs):
        data = {
            'type': 'document',
            'ast': self.get_ast(),
        }

        if (include_metadata):
            data["context"] = self._context

        return data

    def collect_placeholders(self):
        """
        Fetch all the answer placeholders in this document.
        """

        return set(self._collect_placeholders_helper(self._tokens))

    def _collect_placeholders_helper(self, tokens):
        placeholders = []

        if ((tokens is None) or (len(tokens) == 0)):
            return placeholders

        for token in tokens:
            if (token.type == 'placeholder'):
                placeholders.append(token.content)

            placeholders += self._collect_placeholders_helper(token.children)

        return placeholders

    def collect_file_paths(self, base_dir):
        """
        Fetch all the file paths in this document.
        """

        return set(self._collect_file_paths_helper(self._tokens, base_dir))

    def _collect_file_paths_helper(self, tokens, base_dir):
        file_paths = []

        if ((tokens is None) or (len(tokens) == 0)):
            return file_paths

        for token in tokens:
            if (token.type == 'image'):
                src = token.attrGet('src')
                if ((not re.match(r'^http(s)?://', src)) and (not src.startswith('data:image'))):
                    file_paths.append(os.path.realpath(os.path.join(base_dir, src)))

            file_paths += self._collect_file_paths_helper(token.children, base_dir)

        return file_paths

    def is_empty(self):
        return (len(self._tokens) == 0)

    def to_json(self, indent = 4, sort_keys = True, **kwargs):
        return quizcomp.util.json.dumps(self.to_pod(**kwargs), indent = indent, sort_keys = sort_keys)

    def to_format(self, format, **kwargs):
        formatter = getattr(self, 'to_' + format)
        if (formatter is None):
            raise ValueError(f"Unknown format '{format}'.")

        return formatter(**kwargs)

    def get_ast(self):
        """
        Get a represetation of this document's AST.
        """

        return quizcomp.parser.ast.build(self._tokens)



"""
Upload quizes to GradeScope.
"""

import logging
import os
import re
import time

import bs4

import quizcomp.constants
import quizcomp.variant
import quizcomp.util.dirent
import quizcomp.util.httpsession
import quizcomp.util.json

URL_BASE = 'https://www.gradescope.com'
URL_HOMEPAGE = URL_BASE
URL_LOGIN = f"{URL_BASE}/login"
URL_ASSIGNMENTS = f"{URL_BASE}/courses/%s/assignments"
URL_ASSIGNMENT = f"{URL_BASE}/courses/%s/assignments/%s"
URL_ASSIGNMENT_GROUP = f"{URL_BASE}/courses/%s/assignment_containers"
URL_ASSIGNMENT_EDIT = f"{URL_BASE}/courses/%s/assignments/%s/edit"
URL_ASSIGNMENT_RUBRIC = f"{URL_BASE}/courses/%s/assignments/%s/rubric/edit"
URL_ASSIGNMENT_ADD_RUBRIC_ITEM = f"{URL_BASE}/courses/%s/questions/%s/rubric_items"
URL_NEW_ASSIGNMENT_FORM = f"{URL_BASE}/courses/%s/assignments/new"
URL_EDIT_OUTLINE = f"{URL_BASE}/courses/%s/assignments/%s/outline/edit"
URL_PATCH_OUTLINE = f"{URL_BASE}/courses/%s/assignments/%s/outline"

NAME_BOX_ID = 'name'
ID_BOX_ID = 'id'
SIGNATURE_BOX_ID = 'signature'
MANUAL_GRADING_BOX_ID = 'manual_grading'

SPECIAL_QUESTION_TYPES = [
    NAME_BOX_ID,
    ID_BOX_ID,
    SIGNATURE_BOX_ID,
    MANUAL_GRADING_BOX_ID,
]

EXTEND_BOX_QUESTION_TYPES = [
    quizcomp.constants.QUESTION_TYPE_MA,
    quizcomp.constants.QUESTION_TYPE_MCQ,
    quizcomp.constants.QUESTION_TYPE_MDD,
    quizcomp.constants.QUESTION_TYPE_TF,
]

STANDARD_BOX_QUESTION_TYPES = [
    quizcomp.constants.QUESTION_TYPE_ESSAY,
    quizcomp.constants.QUESTION_TYPE_FIMB,
    quizcomp.constants.QUESTION_TYPE_FITB,
    quizcomp.constants.QUESTION_TYPE_MATCHING,
    quizcomp.constants.QUESTION_TYPE_NUMERICAL,
    quizcomp.constants.QUESTION_TYPE_SA,
]

BOX_TYPES = EXTEND_BOX_QUESTION_TYPES + STANDARD_BOX_QUESTION_TYPES + SPECIAL_QUESTION_TYPES

SP_PER_PT = 65536

SESSION_ID_CREATE_ASSIGNMENT_GROUP = 'gradescope_create_assignment_group'
SESSION_ID_UPLOAD = 'gradescope_upload'

class GradeScopeUploader(object):
    def __init__(self, course_id, user, password,
            force = False, rubric = False,
            save_http = False,
            **kwargs):
        super().__init__(**kwargs)

        self.course_id = course_id
        self.user = user
        self.password = password
        self.force = force
        self.rubric = rubric
        self.save_http = save_http

    def upload_quiz(self, variant, base_dir = None, **kwargs):
        """
        Compile a quiz and upload it to GradeScope.

        If base_dir is left None, then a temp dir will be created which will be destroyed on exit.
        If supplied, the base_dir will be left alone when finished.
        """

        if (not isinstance(variant, quizcomp.variant.Variant)):
            raise ValueError("GradeScope quiz uploader requires a quizcomp.variant.Variant type, found %s." % (type(variant)))

        if (base_dir is None):
            base_dir = quizcomp.util.dirent.get_temp_path(prefix = 'quizcomp-gradescope-')

        boxes, special_boxes = self.get_bounding_boxes(variant, base_dir)
        return self.upload(variant, base_dir, boxes, special_boxes)

    def create_assignment_group(self, title, gradescope_ids):
        session = quizcomp.util.httpsession.get_session(SESSION_ID_CREATE_ASSIGNMENT_GROUP, save_http = self.save_http)

        self.login(session)

        assignments_url = URL_ASSIGNMENTS % (self.course_id)
        csrf_token = self.get_csrf_token(session, assignments_url)

        headers = {
            'x-csrf-token': csrf_token,
        }

        post_url = URL_ASSIGNMENT_GROUP % (self. course_id)
        data = {
            'title': title,
            'assignment_ids[]': gradescope_ids,
        }

        response = session.post(post_url, params = data, headers = headers)
        response.raise_for_status()

    def get_bounding_boxes(self, variant, base_dir):
        # {<quetion id>: {<part id>: box, ...}, ...}
        boxes = {}
        # {NAME_BOX_ID: box, ID_BOX_ID: box, SIGNATURE_BOX_ID: box}
        special_boxes = {}

        path = os.path.join(base_dir, "%s.pos" % (variant.title))
        if (not os.path.exists(path)):
            raise ValueError("Could not find path for quiz bounding boxes: '%s'." % (path))

        with open(path, 'r') as file:
            for line in file:
                line = line.strip()
                if (line == ""):
                    continue

                parts = [part.strip() for part in line.split(',')]
                if (len(parts) != 12):
                    raise ValueError("Position file has row with bad number of parts. Expecting 11, found %d." % (len(parts)))

                # "ll" == "lower-left"
                # "ur" == "upper-right"
                (question_id, part_id, answer_id, question_type, page_number, ll_x, ll_y, ur_x, ur_y, page_width, page_height, origin) = parts

                if (origin != 'bottom-left'):
                    raise ValueError("Unknown bounding box origin: '%s'." % (origin))

                # Note that the position file and GradeScope use 1-indexed pages.
                page_number = int(page_number)

                if (question_type not in BOX_TYPES):
                    raise ValueError("Unknown content type: '%s'." % (question_type))

                extend_box_right = False
                if (question_type in EXTEND_BOX_QUESTION_TYPES):
                    extend_box_right = True

                    # There is a special case for inline MA questions.
                    if (question_type == quizcomp.constants.QUESTION_TYPE_MA):
                        extend_box_right = False

                (x1, y1), (x2, y2) = self._compute_box(ll_x, ll_y, ur_x, ur_y, page_width, page_height, extend_box_right = extend_box_right)

                if (question_type in SPECIAL_QUESTION_TYPES):
                    # These boxes are special.
                    if (question_type in special_boxes):
                        raise ValueError("Multiple %s bounding boxes found." % (question_type))

                    special_boxes[question_type] = {
                        'page_number': page_number,
                        'x1': x1,
                        'y1': y1,
                        'x2': x2,
                        'y2': y2,
                    }

                    continue

                if (question_id not in boxes):
                    boxes[question_id] = {}

                # If there is an existing box, extend it.
                if (part_id in boxes[question_id]):
                    old_box = boxes[question_id][part_id]

                    old_x1 = old_box['x1']
                    old_y1 = old_box['y1']
                    old_x2 = old_box['x2']
                    old_y2 = old_box['y2']

                    old_page = old_box['page_number']
                    if (old_page != page_number):
                        raise ValueError("Question %s has bounding boxes that span pages." % (question_id))

                    x1 = min(x1, old_x1)
                    y1 = min(y1, old_y1)
                    x2 = max(x2, old_x2)
                    y2 = max(y2, old_y2)

                boxes[question_id][part_id] = {
                    'page_number': page_number,
                    'x1': x1,
                    'y1': y1,
                    'x2': x2,
                    'y2': y2,
                }

        return boxes, special_boxes

    def _compute_box(self, ll_x, ll_y, ur_x, ur_y, page_width, page_height, extend_box_right = False):
        ll_x = float(ll_x.removesuffix('sp'))
        ll_y = float(ll_y.removesuffix('sp'))
        ur_x = float(ur_x.removesuffix('sp'))
        ur_y = float(ur_y.removesuffix('sp'))

        page_width = float(page_width.removesuffix('pt')) * SP_PER_PT
        page_height = float(page_height.removesuffix('pt')) * SP_PER_PT

        # Origin is upper-left, point 1 is upper-left, point 2 is lower-right.
        x1 = round(100.0 * (ll_x / page_width), 1)
        y1 = round(100.0 * (1.0 - (ur_y / page_height)), 1)
        # The lower right x should always extend (at least) to the end of the page (to capture the answers).
        x2 = round(100.0 * (ur_x / page_width), 1)
        y2 = round(100.0 * (1.0 - (ll_y / page_height)), 1)

        if (extend_box_right):
            # In some question types, we want to extend (at least) to the end of the page (to capture the answers).
            x2 = max(95.0, x2)

        return (x1, y1), (x2, y2)

    def create_outline(self, variant, bounding_boxes, special_boxes):
        question_data = []
        for (question_id, parts) in bounding_boxes.items():
            question_index = int(float(question_id))

            if (len(parts) == 1):
                # Single-part question.
                question_data.append({
                    'title': variant.questions[question_index].name,
                    'weight': variant.questions[question_index].points,
                    'crop_rect_list': list(parts.values()),
                })
            else:
                # Multi-part question.
                children = []
                for (part_id, box) in parts.items():
                    children.append({
                        'title': "%s - %s" % (variant.questions[question_index].name, part_id),
                        'weight': round(variant.questions[question_index].points / len(parts), 2),
                        'crop_rect_list': [box],
                    })

                question_data.append({
                    'title': variant.questions[question_index].name,
                    'weight': variant.questions[question_index].points,
                    # The top-level question just needs one of the bounding boxes.
                    'crop_rect_list': [list(parts.values())[0]],
                    'children': children,
                })

        name_box = None
        id_box = None

        for (question_type, box) in special_boxes.items():
            if (question_type == NAME_BOX_ID):
                name_box = special_boxes[NAME_BOX_ID]
            elif (question_type == ID_BOX_ID):
                id_box = special_boxes[ID_BOX_ID]
            else:
                question_data.append({
                    'title': question_type,
                    'weight': 0,
                    'crop_rect_list': [box],
                })

        outline = {
            'assignment': {
                'identification_regions': {
                    'name': name_box,
                    'sid': id_box
                }
            },
            'question_data': question_data,
        }

        return outline

    def upload(self, variant, base_dir,  bounding_boxes, special_boxes):
        path = os.path.join(base_dir, "%s.pdf" % (variant.title))
        if (not os.path.exists(path)):
            raise ValueError("Could not find path for quiz pdf: '%s'." % (path))

        outline = self.create_outline(variant, bounding_boxes, special_boxes)

        session = quizcomp.util.httpsession.get_session(SESSION_ID_UPLOAD, save_http = self.save_http)

        self.login(session)
        logging.debug("Logged in as '%s'.", self.user)

        assignment_id = self.get_assignment_id(session, variant)
        if (assignment_id is not None):
            if (not self.force):
                logging.info("Assignment '%s' (%s) already exists. Skipping upload.", variant.title, assignment_id)
                return assignment_id, False

            self.delete_assignment(session, assignment_id)
            logging.debug("Deleted assignment: '%s'", assignment_id)

        assignment_id = self.create_assignment(session, variant, base_dir)
        logging.debug("Created assignment: '%s'", assignment_id)

        self.submit_outline(session, assignment_id, outline)
        logging.debug('Submitted outline.')

        if (self.rubric):
            self.create_rubric(session, assignment_id, variant)
            logging.debug('Created assignment rubric.')

        return assignment_id, True

    def login(self, session):
        token = self.get_authenticity_token(session, URL_HOMEPAGE, action = '/login')

        data = {
            'utf8': '✓',
            'session[email]': self.user,
            'session[password]': self.password,
            'session[remember_me]': 0,
            'commit': 'Log+In',
            'session[remember_me_sso]': 0,
            'authenticity_token': token,
        }

        # Login.
        response = session.post(URL_LOGIN, params = data)
        response.raise_for_status()

    def get_authenticity_token(self, session, url, action = None):
        response = session.get(url)
        response.raise_for_status()

        document = bs4.BeautifulSoup(response.text, 'html.parser')

        form_selector = 'form'
        if (action is not None):
            form_selector = 'form[action="%s"]' % (action)

        auth_input = document.select('%s input[name="authenticity_token"]' % (form_selector))
        if (len(auth_input) != 1):
            raise ValueError("Did not find exactly one authentication token input, found %d." % (len(auth_input)))
        auth_input = auth_input[0]

        return auth_input.get('value')

    def get_csrf_token(self, session, url):
        # Get outline submission csrf token.
        response = session.get(url)
        response.raise_for_status()

        document = bs4.BeautifulSoup(response.text, 'html.parser')
        return self.parse_csrf_token(document)

    def parse_csrf_token(self, document):
        meta_tag = document.select('meta[name="csrf-token"]')
        if (len(meta_tag) != 1):
            raise ValueError("Did not find exactly one CSRF meta tag, found %d." % (len(meta_tag)))
        meta_tag = meta_tag[0]

        return meta_tag.get('content')

    def get_assignment_id(self, session, variant):
        url = URL_ASSIGNMENTS % (self.course_id)

        response = session.get(url)
        response.raise_for_status()

        document = bs4.BeautifulSoup(response.text, 'html.parser')

        nodes = document.select('div[data-react-class="AssignmentsTable"]')
        if (len(nodes) != 1):
            raise ValueError("Did not find exactly one assignments table, found %d." % (len(nodes)))

        assignment_data = quizcomp.util.json.loads(nodes[0].get('data-react-props'))

        for row in assignment_data['table_data']:
            if (row['type'] != 'assignment'):
                continue

            id = row['id'].strip().removeprefix('assignment_')
            name = row['title'].strip()

            if (name == variant.title):
                return id

        return None

    def delete_assignment(self, session, assignment_id):
        form_url = URL_ASSIGNMENT_EDIT % (self.course_id, assignment_id)
        token = self.get_csrf_token(session, form_url)

        data = {
            '_method': 'delete',
            'authenticity_token': token,
        }

        delete_url = URL_ASSIGNMENT % (self.course_id, assignment_id)
        response = session.post(delete_url, data = data)
        response.raise_for_status()

    def create_assignment(self, session, variant, base_dir):
        form_url = URL_NEW_ASSIGNMENT_FORM % (self.course_id)
        token = self.get_csrf_token(session, form_url)

        data = {
            'authenticity_token': token,
            'assignment[title]': variant.title,
            'assignment[submissions_anonymized]': 0,
            'assignment[student_submission]': "false",
            'assignment[when_to_create_rubric]': 'while_grading',
            'assignment[scoring_type]': 'negative',
        }

        path = os.path.join(base_dir, "%s.pdf" % (variant.title))
        files = {
            'template_pdf': (
                os.path.basename(path),
                open(path, 'rb')
            ),
        }

        create_url = URL_ASSIGNMENTS % (self.course_id)
        response = session.post(create_url, data = data, files = files)
        response.raise_for_status()

        if (len(response.history) == 0):
            raise ValueError("Failed to create assignment. Is the name ('%s') unique?" % (variant.title))

        match = re.search(r'/assignments/(\d+)/outline/edit', response.history[0].text)
        if (match is None):
            logging.error("--- Create Body ---\n%s\n------" % response.history[0].text)
            raise ValueError("Could not parse assignment ID from response body.")

        return match.group(1)

    def submit_outline(self, session, assignment_id, outline):
        edit_url = URL_EDIT_OUTLINE % (self.course_id, assignment_id)
        csrf_token = self.get_csrf_token(session, edit_url)

        headers = {
            'Content-Type': 'application/json',
            'x-csrf-token': csrf_token,
        }

        patch_outline_url = URL_PATCH_OUTLINE % (self.course_id, assignment_id)
        response = session.patch(patch_outline_url,
            data = quizcomp.util.json.dumps(outline, separators = (',', ':')),
            headers = headers,
        )
        response.raise_for_status()

    def create_rubric(self, session, assignment_id, variant):
        questions_ids, csrf_token = self.fetch_question_ids(session, assignment_id)

        for question in variant.questions:
            if (question.name not in questions_ids):
                continue

            question_ids = questions_ids[question.name]
            score = round(question.points / len(question_ids), 2)

            for question_id in question_ids:
                self.add_rubric_item(session, csrf_token, question_id, "Incorrect", score)

    def add_rubric_item(self, session, csrf_token, question_id, description, score):
        url = URL_ASSIGNMENT_ADD_RUBRIC_ITEM % (self.course_id, question_id)

        headers = {
            'Content-Type': 'application/json',
            'x-csrf-token': csrf_token,
        }

        data = {
            "rubric_item":{
                "description": description,
                "weight": str(score),
                "group_id": None,
            },
        }

        response = session.post(url, json = data, headers = headers)
        response.raise_for_status()

    def fetch_question_ids(self, session, assignment_id):
        """
        Get the GradeScope id of all questions/subquestions.

        Returns: {<question base name>: [question id, ...], ...}
        """

        url = URL_ASSIGNMENT_RUBRIC % (self.course_id, assignment_id)

        response = session.get(url)
        response.raise_for_status()

        document = bs4.BeautifulSoup(response.text, 'html.parser')

        csrf_token = self.parse_csrf_token(document)

        data_tag = document.select('div[data-react-class="AssignmentRubric"]')
        if (len(data_tag) != 1):
            raise ValueError("Did not find exactly one rubric data tag, found %d." % (len(data_tag)))
        data_tag = data_tag[0]

        data = quizcomp.util.json.loads(data_tag.get('data-react-props'))

        ids = {}
        for question in data['questions']:
            question_ids = []

            children = question.get('children', None)
            if (children is None):
                question_ids.append(str(question['id']))
            else:
                question_ids = [str(child['id']) for child in children]

            ids[question['title']] = question_ids

        return ids, csrf_token



"""
The `quizcomp.cli.gradescope` package contains tools for interacting with Gradescope.
"""

import sys

import quizcomp.util.cli

def main():
    return quizcomp.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())



import quizcomp.common
import quizcomp.constants
import quizcomp.question.base

class TextOnly(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_TEXT_ONLY):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        if (self.answers is not None):
            raise quizcomp.common.QuestionValidationError(self, "'answers' key must be missing or None/null, found: '%s'." % (
                    str(self.answers)))



"""
Read (an optionally save) a Quiz Composer project.
"""

import sys

import quizcomp.args
import quizcomp.project

def run(args):
    project = quizcomp.project.Project.from_path(args.path)

    quizzes, questions = project.load_resources()

    print("Found %d quizzes." % (len(quizzes)))
    for (path, quiz) in quizzes:
        print("    %s (%s)" % (path, quiz.title))

    print("Found %d question." % (len(questions)))
    for (path, question) in questions:
        text = "    %s" % (path)
        if (question.name != ''):
            text += " (%s)" % (question.name)

        print(text)

    if (args.out_dir is not None):
        project.save(args.out_dir)

    return 0

def _get_parser():
    parser = quizcomp.args.Parser(description =
        __doc__.strip())

    parser.add_argument('path', metavar = 'PATH',
        type = str,
        help = 'The path to a project dir or config file.')

    parser.add_argument('--outdir', dest = 'out_dir',
        action = 'store', type = str, default = None,
        help = 'Save the project to this directory.')

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import abc

import quizcomp.variant

class Converter(abc.ABC):
    def __init__(self, answer_key = False, **kwargs):
        super().__init__()

        self.answer_key = answer_key

    @abc.abstractmethod
    def convert_variant(self, variant, **kwargs):
        pass

    def convert_question(self, question, **kwargs):
        """
        Convert a single question using a dummy quiz layout.
        """

        variant = quizcomp.variant.Variant.get_dummy(question)
        return self.convert_variant(variant, **kwargs)



import datetime
import logging
import os
import random

import quizcomp.common
import quizcomp.constants
import quizcomp.group
import quizcomp.parser.public
import quizcomp.uploader.canvas
import quizcomp.util.dirent
import quizcomp.util.git
import quizcomp.util.serial

class Quiz(quizcomp.util.serial.JSONSerializer):
    """
    A quiz object represents multiple possible assesments (called "variants").
    """

    def __init__(self, type = quizcomp.constants.TYPE_QUIZ,
            title = '',
            course_title = '', term_title = '',
            description = '', date = '',
            time_limit_mins = None,
            shuffle_answers = True, pick_with_replacement = True,
            groups = [],
            base_dir = '.',
            version = None, seed = None,
            canvas = {},
            ids = {},
            **kwargs):
        super().__init__(**kwargs)

        self.title = title
        self.course_title = course_title
        self.term_title = term_title
        self.date = date

        self.description = description

        self.time_limit_mins = time_limit_mins

        self.shuffle_answers = shuffle_answers
        self.pick_with_replacement = pick_with_replacement

        self.groups = groups

        self.base_dir = base_dir
        self.version = version

        self.seed = seed
        if (self.seed is None):
            self.seed = random.randint(0, 2**64)
        self._rng = random.Random(self.seed)

        self.canvas = canvas.copy()

        try:
            self.validate(cls = Quiz, **kwargs)
        except Exception as ex:
            ids = ids.copy()
            ids['title'] = self.title

            raise quizcomp.common.QuizValidationError('Error while validating quiz.', ids = ids) from ex

    def _validate(self, **kwargs):
        if ((self.title is None) or (self.title == "")):
            raise quizcomp.common.QuizValidationError("Title cannot be empty.")

        if ((self.description is None) or (self.description == "")):
            raise quizcomp.common.QuizValidationError("Description cannot be empty.")
        self.description = quizcomp.parser.public.parse_text(self.description, base_dir = self.base_dir)

        if (self.version is None):
            self.version = quizcomp.util.git.get_version(self.base_dir, throw = False)
            if (self.version == quizcomp.util.git.UNKNOWN_VERSION):
                logging.warning("Could not get a version for the quiz (is it in a git repo?).")

        self.canvas = quizcomp.uploader.canvas.validate_options(self.canvas)

        self._validate_time_limit()

        if (self.date == ''):
            self.date = datetime.date.today()
        elif (isinstance(self.date, str)):
            self.date = datetime.date.fromisoformat(self.date)
        else:
            raise quizcomp.common.QuizValidationError("Date should be a string or datetime.date, found '%s'." % (str(type(self.date))))

        for key in kwargs:
            logging.warning("Unknown quiz option: '%s'." % (key))

    def _validate_time_limit(self):
        if (self.time_limit_mins is None):
            return

        if (not isinstance(self.time_limit_mins, (str, int))):
            raise quizcomp.common.QuizValidationError("Time limit must be a positive int, found '%s'." % (str(self.time_limit_mins)))

        try:
            self.time_limit_mins = int(self.time_limit_mins)
        except:
            raise quizcomp.common.QuizValidationError("Time limit must be a positive int, found '%s'." % (str(self.time_limit_mins)))

        if (self.time_limit_mins < 0):
            raise quizcomp.common.QuizValidationError("Time limit must be a positive int, found '%s'." % (str(self.time_limit_mins)))

        if (self.time_limit_mins == 0):
            self.time_limit_mins = None

    @classmethod
    def from_path(cls, path, **kwargs):
        # Check for a description file.
        def _check_description_file(path, data):
            description_filename = os.path.splitext(os.path.basename(path))[0]
            description_path = os.path.join(os.path.dirname(path), description_filename + '.md')
            if (os.path.exists(description_path)):
                data['description'] = quizcomp.util.dirent.read_file(description_path)
                logging.debug("Loading quiz description from '%s'.", description_path)

            return data

        return super().from_path(path, data_callback = _check_description_file, **kwargs)

    @staticmethod
    def from_dict(quiz_info, base_dir = None, flatten_groups = False, ids = {}, **kwargs):
        groups = []
        group_infos = quiz_info.get('groups', [])
        for i in range(len(group_infos)):
            ids = ids.copy()
            ids['index'] = i
            groups.append(quizcomp.group.Group.from_dict(group_infos[i], base_dir, ids = ids))

        if (flatten_groups):
            new_groups = []

            for old_group in groups:
                for i in range(len(old_group.questions)):
                    info = {
                        'name': old_group.name,
                        'pick_count': 1,
                        'points': old_group.points,
                        'questions': [old_group.questions[i]],
                    }

                    new_groups.append(quizcomp.group.Group(**info))

            groups = new_groups

        quiz_info['groups'] = groups

        if (base_dir is not None):
            quiz_info['base_dir'] = base_dir
        elif ('base_dir' not in quiz_info):
            quiz_info['base_dir'] = '.'

        return Quiz(**quiz_info, ids = ids)

    def num_questions(self):
        count = 0

        for group in self.groups:
            count += group.pick_count

        return count

    def create_variant(self, identifier = None, seed = None, all_questions = False):
        if (seed is None):
            seed = self._rng.randint(0, 2**64)

        logging.debug("Creating variant with seed %s.", str(seed))
        rng = random.Random(seed)

        new_groups = []
        for group in self.groups:
            questions = group.choose_questions(all_questions = all_questions, rng = rng,
                    with_replacement = self.pick_with_replacement)

            group_data = group.__dict__.copy()
            group_data['questions'] = questions
            # Skip validation.
            group_data['_skip_all_validation'] = True

            new_groups.append(quizcomp.group.Group(**group_data))

        if (self.shuffle_answers):
            for group in new_groups:
                for question in group.questions:
                    question.shuffle(rng)

        title = self.title
        version = self.version

        if (identifier is not None):
            title = "%s - %s" % (title, identifier)
            version = "%s, Variant: %s" % (version, identifier)

        data = self.__dict__.copy()

        data['title'] = title
        data['version'] = version
        data['seed'] = seed
        data['groups'] = new_groups

        # Skip quiz validation.
        data['_skip_class_validations'] = [Quiz]

        return quizcomp.variant.Variant(**data)



"""
Style constants and base functionality.
"""

KEY_CONTENT_ALIGN = 'content-align'
KEY_FONT_SIZE = 'font-size'
KEY_IMAGE_WIDTH = 'image-width'
KEY_TABLE_BORDER_CELLS = 'table-border-cells'
KEY_TABLE_BORDER_TABLE = 'table-border-table'
KEY_TABLE_CELL_HEIGHT = 'table-cell-height'
KEY_TABLE_CELL_WIDTH = 'table-cell-width'
KEY_TABLE_HEAD_BOLD = 'table-head-bold'
KEY_TABLE_HEAD_RULE = 'table-head-rule'
KEY_TEXT_ALIGN = 'text-align'

DEFAULT_IMAGE_WIDTH = 1.0
DEFAULT_TABLE_BORDER_CELLS = False
DEFAULT_TABLE_BORDER_TABLE = False
DEFAULT_TABLE_CELL_HEIGHT = 1.5
DEFAULT_TABLE_CELL_WIDTH = 1.5
DEFAULT_TABLE_HEAD_BOLD = True
DEFAULT_TABLE_HEAD_RULE = True

ALLOWED_VALUES_ALIGNMENT_LEFT = 'left'
ALLOWED_VALUES_ALIGNMENT_CENTER = 'center'
ALLOWED_VALUES_ALIGNMENT_RIGHT = 'right'
ALLOWED_VALUES_ALIGNMENT = [
    ALLOWED_VALUES_ALIGNMENT_LEFT,
    ALLOWED_VALUES_ALIGNMENT_CENTER,
    ALLOWED_VALUES_ALIGNMENT_RIGHT
]

FLEXBOX_ALIGNMENT = {
    ALLOWED_VALUES_ALIGNMENT_LEFT: 'flex-start',
    ALLOWED_VALUES_ALIGNMENT_CENTER: 'center',
    ALLOWED_VALUES_ALIGNMENT_RIGHT: 'flex-end',
}

TEX_BLOCK_ALIGNMENT = {
    ALLOWED_VALUES_ALIGNMENT_LEFT: 'flushleft',
    ALLOWED_VALUES_ALIGNMENT_CENTER: 'center',
    ALLOWED_VALUES_ALIGNMENT_RIGHT: 'flushright',
}

def get_alignment(style, key, default_value = None):
    alignment = style.get(key, None)
    if (alignment is None):
        return default_value

    alignment = str(alignment).lower()
    if (alignment not in ALLOWED_VALUES_ALIGNMENT):
        raise ValueError("Unknown value for '%s' style key '%s'. Allowed values: '%s'." % (key, alignment, ALLOWED_VALUES_ALIGNMENT))

    return alignment

def get_boolean_style_key(style, key, default_value = None):
    value = style.get(key, default_value)
    if (value is None):
        return default_value

    return (value is True)

def get_image_width(style):
    width = style.get(KEY_IMAGE_WIDTH, None)
    if (width is None):
        width = DEFAULT_IMAGE_WIDTH

    return float(width)

def compute_html_style_string(style):
    """
    Compute the attribute style string for an HTML tag.
    """

    attributes = []

    content_align = get_alignment(style, KEY_CONTENT_ALIGN)
    if (content_align is not None):
        attributes.append("display: flex")
        attributes.append("flex-direction: column")
        attributes.append("justify-content: flex-start")
        attributes.append("align-items: %s" % (FLEXBOX_ALIGNMENT[content_align]))

    text_align = get_alignment(style, KEY_TEXT_ALIGN)
    if (text_align is not None):
        attributes.append("text-align: %s" % (text_align))

    font_size = style.get(KEY_FONT_SIZE, None)
    if (font_size is not None):
        attributes.append("font-size: %.2fpt" % (float(font_size)))

    if (len(attributes) == 0):
        return ''

    return '; '.join(attributes)

def compute_tex_fixes(style):
    """
    Compute the fixes (prefixes, suffixes) for a portion of TeX.
    These are things like `\\begin{center}`/`\\end{center}`.
    The returned lists have matching indexes.
    """

    # The beginning and ends of groups.
    # These will match 1-1.
    prefixes = []
    suffixes = []

    content_align = get_alignment(style, KEY_CONTENT_ALIGN)
    if (content_align is not None):
        env_name = TEX_BLOCK_ALIGNMENT[content_align]
        prefixes.append("\\begin{%s}" % (env_name))
        suffixes.append("\\end{%s}" % (env_name))

    font_size = style.get(KEY_FONT_SIZE, None)
    if (font_size is not None):
        font_size = float(font_size)
        # 1.2 is the default size for baseline skip relative to font size.
        # See: https://ctan.math.illinois.edu/macros/latex/contrib/fontsize/fontsize.pdf
        baseline_skip = 1.2 * font_size

        prefixes.append('\\begingroup\\fontsize{%.2fpt}{%.2fpt}\\selectfont' % (font_size, baseline_skip))
        suffixes.append('\\endgroup')

    return prefixes, suffixes



TYPE_PROJECT = 'project'
TYPE_QUIZ = 'quiz'
TYPE_QUESTION = 'question'
TYPE_VARIANT = 'variant'

JSON_OBJECT_TYPES = [TYPE_PROJECT, TYPE_QUIZ, TYPE_QUESTION, TYPE_VARIANT]

PROJECT_FILENAME = 'project.json'
QUIZ_FILENAME = 'quiz.json'
QUESTION_FILENAME = 'question.json'
PROMPT_FILENAME = 'prompt.md'

FORMAT_CANVAS = 'canvas'
FORMAT_HTML = 'html'
FORMAT_JSON = 'json'
FORMAT_MD = 'md'
FORMAT_TEX = 'tex'
FORMAT_TEXT = 'text'
FORMAT_QTI = 'qti'

# A special format mainly for testing.
FORMAT_JSON_TEMPLATE = 'json_template'

# Output formats.
DOC_FORMATS = [FORMAT_CANVAS, FORMAT_HTML, FORMAT_JSON, FORMAT_TEX, FORMAT_QTI]

# Formats that the parser understands basic conversions to.
PARSER_FORMATS = [FORMAT_CANVAS, FORMAT_HTML, FORMAT_JSON, FORMAT_MD, FORMAT_TEX, FORMAT_TEXT]

QUESTION_TYPE_ESSAY = 'essay'
QUESTION_TYPE_FIMB = 'fill_in_multiple_blanks'
QUESTION_TYPE_FITB = 'fill_in_the_blank'
QUESTION_TYPE_MATCHING = 'matching'
QUESTION_TYPE_MA = 'multiple_answers'
QUESTION_TYPE_MCQ = 'multiple_choice'
QUESTION_TYPE_MDD = 'multiple_dropdowns'
QUESTION_TYPE_NUMERICAL = 'numerical'
QUESTION_TYPE_SA = 'short_answer'
QUESTION_TYPE_TEXT_ONLY = 'text_only'
QUESTION_TYPE_TF = 'true_false'

# Supported question types.
QUESTION_TYPES = [
    QUESTION_TYPE_ESSAY,
    QUESTION_TYPE_FIMB,
    QUESTION_TYPE_FITB,
    QUESTION_TYPE_MATCHING,
    QUESTION_TYPE_MA,
    QUESTION_TYPE_MCQ,
    QUESTION_TYPE_MDD,
    QUESTION_TYPE_NUMERICAL,
    QUESTION_TYPE_SA,
    QUESTION_TYPE_TEXT_ONLY,
    QUESTION_TYPE_TF,
]

NUMERICAL_ANSWER_TYPE_EXACT = 'exact'
NUMERICAL_ANSWER_TYPE_RANGE = 'range'
NUMERICAL_ANSWER_TYPE_PRECISION = 'precision'



import sys

import quizcomp.args
import quizcomp.constants
import quizcomp.parser.public

def run(args):
    document = quizcomp.parser.public.parse_file(args.path).document

    content = document.to_format(args.format, pretty = True)
    print(content)

    return 0

def _get_parser():
    parser = quizcomp.args.Parser(description =
        "Parse a single file and output the results of the parse.")

    parser.add_argument('path', metavar = 'PATH',
        type = str,
        help = 'The path to parse.')

    parser.add_argument('--format',
        action = 'store', type = str, default = quizcomp.constants.FORMAT_JSON,
        choices = quizcomp.constants.PARSER_FORMATS,
        help = 'Output the parsed document in this format (default: %(default)s).')

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import math
import os
import random
import re
import string

import quizcomp.constants
import quizcomp.converter.converter
import quizcomp.parser.public
import quizcomp.quiz
import quizcomp.util.dirent
import quizcomp.util.http
import quizcomp.variant

import jinja2

TEMPLATE_FILENAME_QUIZ = 'quiz.template'
TEMPLATE_FILENAME_QUESTION_SEP = 'question-separator.template'
TEMPLATE_FILENAME_GROUP = 'group.template'

RIGHT_IDS = string.ascii_uppercase
LEFT_IDS = [str(i + 1) for i in range(len(RIGHT_IDS))]

DEFAULT_JINJA_OPTIONS = {
    'trim_blocks': True,
    'lstrip_blocks': True,
    'autoescape': jinja2.select_autoescape(),
}

DEFAULT_ID_DELIM = '.'

class TemplateConverter(quizcomp.converter.converter.Converter):
    def __init__(self, format, template_dir,
            jinja_options = {}, jinja_filters = {}, jinja_globals = {},
            parser_format_options = {},
            image_base_dir = None, image_relative_root = None, cleanup_images = True,
            id_delim = DEFAULT_ID_DELIM,
            **kwargs):
        super().__init__(**kwargs)

        if (not os.path.isdir(template_dir)):
            raise ValueError("Provided template dir ('%s') does not exist or is not a dir." % (
                template_dir))

        self.format = format
        self.template_dir = template_dir

        self.parser_format_options = parser_format_options
        self.id_delim = id_delim

        # Some converters will need to store image paths.
        # Using the _store_images() callback will put the images inside image_base_dir.
        self.image_base_dir = image_base_dir

        # This will hold: {<abs_path or link>: <new path (based on image_base_dir)>, ...}
        self.image_paths = {}

        # If not None, override the default image output path with os.join(image_relative_root, filename).
        self.image_relative_root = image_relative_root

        # Remove any temp image directories.
        self.cleanup_images = cleanup_images

        self.jinja_options = DEFAULT_JINJA_OPTIONS.copy()
        self.jinja_options.update(jinja_options)

        self.env = jinja2.Environment(
            loader = jinja2.FileSystemLoader(self.template_dir, followlinks = True),
            **self.jinja_options,
        )

        self.env.globals.update(jinja_globals)

        for (name, function) in jinja_filters.items():
            self.env.filters[name] = function

        # Methods to generate answers.
        # Signature: func(self, question_id, question_number, question, variant)
        self.answer_functions = {
            quizcomp.constants.QUESTION_TYPE_ESSAY: 'create_answers_essay',
            quizcomp.constants.QUESTION_TYPE_FIMB: 'create_answers_fimb',
            quizcomp.constants.QUESTION_TYPE_FITB: 'create_answers_fitb',
            quizcomp.constants.QUESTION_TYPE_MA: 'create_answers_ma',
            quizcomp.constants.QUESTION_TYPE_MATCHING: 'create_answers_matching',
            quizcomp.constants.QUESTION_TYPE_MCQ: 'create_answers_mcq',
            quizcomp.constants.QUESTION_TYPE_MDD: 'create_answers_mdd',
            quizcomp.constants.QUESTION_TYPE_NUMERICAL: 'create_answers_numerical',
            quizcomp.constants.QUESTION_TYPE_SA: 'create_answers_sa',
            quizcomp.constants.QUESTION_TYPE_TEXT_ONLY: 'create_answers_text_only',
            quizcomp.constants.QUESTION_TYPE_TF: 'create_answers_tf',
        }

    def convert_quiz(self, quiz, **kwargs):
        return self._convert_container(quiz, quizcomp.quiz.Quiz, 'quiz')

    def convert_variant(self, variant, **kwargs):
        return self._convert_container(variant, quizcomp.variant.Variant, 'variant')

    def _convert_container(self, container, container_type, container_label):
        if (not isinstance(container, container_type)):
            raise ValueError("Template %s converter requires a %s, found %s." % (
                    container_label, str(container_type), type(container)))

        _, inner_text = self.create_groups(container)

        inner_context = container.to_dict()
        inner_context['description_text'] = self._format_doc(container.description.document)

        context = {
            'quiz': inner_context,
            'answer_key': self.answer_key,
            'inner_text': inner_text,
        }

        template = self.env.get_template(TEMPLATE_FILENAME_QUIZ)
        text = template.render(**context)

        return text

    def create_groups(self, quiz):
        return self._create_item_collection(quiz, 'groups', 'group', 1, self.create_group)

    def _create_item_collection(self, container, container_attr, label, question_number, item_creation_function, id_prefix = None):
        """
        Create a collection of groups (for quizzes) or questions (for variants or inside groups)
        from a container (variant or quiz).
        """

        result = []
        items = getattr(container, container_attr)

        for index in range(len(items)):
            item = items[index]

            item_id = str(index)
            if (id_prefix is not None):
                item_id = self.id_delim.join([id_prefix, item_id])

            if (index != 0):
                result.append(self.create_question_separator(container))

            try:
                question_number, text = item_creation_function(item_id, question_number, item, container)
                result.append(text)
            except Exception as ex:
                raise ValueError("Failed to convert %s %d (%s: %s)." % (label, index, item_id, item.name)) from ex

        return question_number, "\n\n".join(result)

    def create_group(self, group_index, question_number, group, quiz):
        data = group.to_dict()
        data['id'] = group_index

        question_number, questions_text = self._create_item_collection(group, 'questions', 'question', question_number, self.create_question, id_prefix = group_index)

        context = {
            'quiz': quiz,
            'group': data,
            'questions_text': questions_text,
        }

        template = self.env.get_template(TEMPLATE_FILENAME_GROUP)
        text = template.render(**context)

        return question_number, text

    def create_question(self, question_id, question_number, question, variant):
        question_type = question.question_type
        if (question_type not in self.answer_functions):
            raise ValueError("Unsupported question type: '%s'." % (question_type))

        data = question.to_dict()
        data['prompt_text'] = self._format_doc(question.prompt.document)
        data['id'] = question_id
        data['number'] = question_number

        # Stash the old answers and add in new ones.
        data['answers_raw'] = data['answers']
        answers_method = getattr(self, self.answer_functions[question_type])
        data['answers'] = answers_method(question_id, question_number, question, variant)

        data['feedback'] = {}
        for key, item in question.feedback.items():
            data['feedback'][key] = self._format_doc(item.document)

        context = {
            'quiz': variant,
            'answer_key': self.answer_key,
            'question': data,
        }

        template_name = "%s.template" % (question_type)
        template = self.env.get_template(template_name)

        context = self.modify_question_context(context, question, variant)
        text = template.render(**context)

        if (not question.should_skip_numbering()):
            question_number += 1

        return question_number, text

    def modify_question_context(self, context, question, variant):
        """
        Provide an opportunity for children to modify the question context.
        The new context reference (which may be new, unchanged, or modified version of the passed-in context).
        """

        return context

    def clean_solution_content(self, document):
        """
        An opportunity for children to clean the text of a solution before it is entered into a key.
        For example, tex solutions are hacky and cannot use certain functions.
        """

        return self._format_doc(document)

    def create_question_separator(self, variant):
        context = {
            'quiz': variant,
            'answer_key': self.answer_key,
        }

        template = self.env.get_template(TEMPLATE_FILENAME_QUESTION_SEP)
        text = template.render(**context)

        return text

    def create_answers_tf(self, question_id, question_number, question, variant):
        return question.answers

    def create_answers_matching(self, question_id, question_number, question, variant):
        lefts = []
        rights = []

        # {left_index: right_index, ...}
        matches = {}

        for items in question.answers['matches']:
            matches[len(lefts)] = len(rights)

            lefts.append({
                'initial_text': items['left'].text,
                'raw_text': self._format_doc(items['left'].document, doc_format = quizcomp.constants.FORMAT_TEXT),
                'text': self._format_doc(items['left'].document),
            })

            rights.append({
                'initial_text': items['right'].text,
                'raw_text': self._format_doc(items['right'].document, doc_format = quizcomp.constants.FORMAT_TEXT),
                'text': self._format_doc(items['right'].document),
            })

        for right in question.answers['distractors']:
            rights.append({
                'initial_text': right.text,
                'raw_text': self._format_doc(right.document, doc_format = quizcomp.constants.FORMAT_TEXT),
                'text': self._format_doc(right.document),
            })

        left_ids = self.get_matching_left_ids()
        right_ids = self.get_matching_right_ids()

        if (len(lefts) > len(left_ids)):
            raise ValueError("Too many left-hand values for a matching question. Found: %d, Max %d." % (len(lefts) > len(left_ids)))

        if (len(rights) > len(right_ids)):
            raise ValueError("Too many right-hand values for a matching question. Found: %d, Max %d." % (len(rights) > len(right_ids)))

        if (question.answers.get('shuffle', False)):
            seed = question.answers.get('shuffle_seed', None)
            rng = random.Random(seed)

            # Shuffle the left and right options while maintining the match mapping.
            left_indexes = list(range(len(lefts)))
            right_indexes = list(range(len(rights)))

            rng.shuffle(left_indexes)
            rng.shuffle(right_indexes)

            new_lefts = [lefts[index] for index in left_indexes]
            lefts = new_lefts

            new_rights = [rights[index] for index in right_indexes]
            rights = new_rights

            new_matches = {left_indexes.index(old_left_index): right_indexes.index(old_right_index) for (old_left_index, old_right_index) in matches.items()}
            matches = new_matches

        # Augment the left and rights with more information for the template.
        for left_index in range(len(lefts)):
            right_index = matches[left_index]

            lefts[left_index] = {
                'id': self.id_delim.join([question_id, left_ids[left_index]]),
                'text': lefts[left_index]['text'],
                'initial_text': lefts[left_index]['initial_text'],
                'raw_text': lefts[left_index]['raw_text'],
                'solution': right_ids[right_index],
                'solution_id': self.id_delim.join([question_id, right_ids[right_index]]),
                'index': left_index,
                'solution_index': right_index,
                'one_index': left_index + 1,
                'solution_one_index': right_index + 1,
            }

        for right_index in range(len(rights)):
            rights[right_index] = {
                'id': self.id_delim.join([question_id, right_ids[right_index]]),
                'text': rights[right_index]['text'],
                'initial_text': rights[right_index]['initial_text'],
                'raw_text': rights[right_index]['raw_text'],
                'label': right_ids[right_index],
                'index': right_index,
                'one_index': right_index + 1,
            }

        return {
            'lefts': lefts,
            'rights': rights,
            'matches': matches,
        }

    def get_matching_left_ids(self):
        return LEFT_IDS

    def get_matching_right_ids(self):
        return RIGHT_IDS

    def create_answers_mcq(self, question_id, question_number, question, variant):
        return self._create_answers_mcq_list(question.answers)

    def _create_answers_mcq_list(self, answers):
        choices = []

        for i in range(len(answers)):
            answer = answers[i]
            choice = self._create_answers_text_value(answer)
            choice['correct'] = answer.is_correct()

            choices.append(choice)

        return choices

    def create_answers_text_only(self, question_id, question_number, question, variant):
        return None

    def create_answers_numerical(self, question_id, question_number, question, variant):
        answer = question.answers[0]

        if (answer.type == quizcomp.constants.NUMERICAL_ANSWER_TYPE_EXACT):
            if (math.isclose(answer.margin, 0.0)):
                content = "%s" % (str(answer.value))
            else:
                content = "%s ± %f" % (str(answer.value), answer.margin)
        elif (answer.type == quizcomp.constants.NUMERICAL_ANSWER_TYPE_RANGE):
            content = "[%s, %s]" % (str(answer.min), str(answer.max))
        elif (answer.type == quizcomp.constants.NUMERICAL_ANSWER_TYPE_PRECISION):
            content = "%s (precision: %s)" % (str(answer.value), str(answer.precision))
        else:
            raise ValueError(f"Unknown numerical answer type: '{answer.type}'.")

        document = quizcomp.parser.public.parse_text(content).document

        return {
            'solution': self.clean_solution_content(document),
            'dirty_solution': self._format_doc(document),
            'raw_solution': self._format_doc(document, doc_format = quizcomp.constants.FORMAT_TEXT),
            'raw_answers': question.answers,
        }

    def create_answers_mdd(self, question_id, question_number, question, variant):
        answers = []

        for key, items in question.answers.items():
            answers.append({
                'label': self._format_doc(items['key'].document),
                'initial_label': items['key'].text,
                'raw_label': self._format_doc(items['key'].document, doc_format = quizcomp.constants.FORMAT_TEXT),
                'choices': self._create_answers_mcq_list(items['values']),
            })

        return answers

    def create_answers_ma(self, question_id, question_number, question, variant):
        return self._create_answers_mcq_list(question.answers)

    def create_answers_fimb(self, question_id, question_number, question, variant):
        answers = {}

        for (key, item) in question.answers.items():
            solutions = []
            for value in item['values']:
                solutions.append(self._create_answers_text_value(value))

            answers[key] = {
                'label': self._format_doc(item['key'].document),
                'raw_label': self._format_doc(item['key'].document, doc_format = quizcomp.constants.FORMAT_TEXT),
                'initial_label': item['key'].text,
                'solutions': solutions,
            }

        return answers

    def create_answers_fitb(self, question_id, question_number, question, variant):
        return self.create_answers_fimb(question_id, question_number, question, variant)['']['solutions']

    def create_answers_sa(self, question_id, question_number, question, variant):
        return self._create_answers_text(question_id, question_number, question, variant)

    def create_answers_essay(self, question_id, question_number, question, variant):
        return self._create_answers_text(question_id, question_number, question, variant)

    def _create_answers_text(self, question_id, question_number, question, variant):
        solutions = []
        for value in question.answers:
            solutions.append(self._create_answers_text_value(value))

        return solutions

    def _create_answers_text_value(self, value):
        """
        Create an output dict for a value that was parsed from text (the result of a parsed string).
        """

        result = {
            'text': self._format_doc(value.document),
            'raw_text': self._format_doc(value.document, doc_format = quizcomp.constants.FORMAT_TEXT),
            'initial_text': value.text,
            'clean': self.clean_solution_content(value.document),
        }

        if (value.has_feedback()):
            result.update({
                'feedback': self._format_doc(value.feedback.document),
                'raw_feedback': self._format_doc(value.feedback.document, doc_format = quizcomp.constants.FORMAT_TEXT),
                'initial_feedback': value.feedback.text,
            })

        return result

    def _format_doc(self, doc, doc_format = None, format_options = None):
        if (doc_format is None):
            doc_format = self.format

        if (format_options is None):
            format_options = self.parser_format_options

        return doc.to_format(doc_format, **format_options)

    def _store_images(self, link, base_dir):
        if (self.image_base_dir is None):
            self.image_base_dir = quizcomp.util.dirent.get_temp_path(prefix = 'quizcomp-images-', rm = self.cleanup_images)

        os.makedirs(self.image_base_dir, exist_ok = True)

        if (re.match(r'^http(s)?://', link)):
            temp_dir = quizcomp.util.dirent.get_temp_path(prefix = 'quizcomp-image-dl-')
            in_path = quizcomp.util.http.get_file(link, temp_dir)
            image_id = link
        else:
            in_path = os.path.join(base_dir, link)
            image_id = in_path

        ext = os.path.splitext(in_path)[-1]
        filename = "%03d%s" % (len(self.image_paths), ext)
        out_path = os.path.join(self.image_base_dir, filename)

        quizcomp.util.dirent.copy_dirent(in_path, out_path)

        if (self.image_relative_root is not None):
            out_path = os.path.join(self.image_relative_root, filename)

        self.image_paths[image_id] = out_path

        return out_path



"""
Convert a quiz to JSON.
"""

import os

import quizcomp.constants
import quizcomp.converter.converter
import quizcomp.converter.template

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DEFAULT_TEMPLATE_DIR = os.path.join(THIS_DIR, '..', 'data', 'templates', 'edq-json')

class JSONTemplateConverter(quizcomp.converter.template.TemplateConverter):
    def __init__(self,
            format = quizcomp.constants.FORMAT_JSON, template_dir = DEFAULT_TEMPLATE_DIR,
            **kwargs):
        super().__init__(format, template_dir, **kwargs)

    # Simplify parts of the question context (specifically the answers) for testing.
    def modify_question_context(self, context, question, variant):
        question = context['question']
        answers = question['answers']
        question_type = question['question_type']

        question['answers'] = self._clean_answers(answers, question_type)

        return context

    def _clean_answers(self, answers, question_type):
        if (answers is None):
            # Seen in text only questions.
            return None

        if (isinstance(answers, list)):
            return self._clean_answers_list(answers, question_type)

        if (question_type == quizcomp.constants.QUESTION_TYPE_MATCHING):
            return self._clean_answers_matching(answers)

        if (question_type == quizcomp.constants.QUESTION_TYPE_NUMERICAL):
            return [raw_answer.to_pod() for raw_answer in answers['raw_answers']]

        if (question_type == quizcomp.constants.QUESTION_TYPE_FIMB):
            result = []
            for answer in answers.values():
                result.append({
                    'label': answer['raw_label'],
                    'solutions': [value['raw_text'] for value in answer['solutions']],
                })

            return result


        raise ValueError(f"Unknown answers type: '{type(answers)}'.")

    def _clean_answers_list(self, answers, question_type):
        for i in range(len(answers)):
            old_answer = answers[i]

            if (not isinstance(old_answer, dict)):
                old_answer = old_answer.to_pod()

            new_answer = {}

            if (question_type == quizcomp.constants.QUESTION_TYPE_MDD):
                new_answer['label'] = old_answer['raw_label']
                new_answer['choices'] = self._clean_answers_list(old_answer['choices'], None)
            else:
                text_keys = ['raw_text', 'text']
                for key in text_keys:
                    if (key in old_answer):
                        new_answer['text'] = old_answer[key]
                        break

                if ('correct' in old_answer):
                    new_answer['correct'] = old_answer['correct']

            answers[i] = new_answer

        return answers

    def _clean_answers_matching(self, answers):
        return {
            'lefts': [self._clean_matching_item(item) for item in answers['lefts']],
            'rights': [self._clean_matching_item(item) for item in answers['rights']],
            'distractors': [item['raw_text'] for item in answers.get('distractors', [])],
        }

    def _clean_matching_item(self, item):
        result = {
            'text': item['raw_text'],
            'id': item['id'],
        }

        if ('solution_id' in item):
            result['solution_id'] = item['solution_id']

        return result

class JSONConverter(quizcomp.converter.converter.Converter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def convert_variant(self, variant, **kwargs):
        return variant.to_json()



import copy
import os
import re
import types

import markdown_it.renderer

import quizcomp.constants
import quizcomp.parser.common
import quizcomp.parser.image
import quizcomp.parser.math
import quizcomp.parser.style

HTML_BORDER_SPEC = '1px solid black'

ADDITIONAL_STYLE = {
    'paragraph_open': [
        'margin-top: 0',
    ],
    'fence_open': [
        'margin-top: 0',
    ],
    'code_block_open': [
        'margin-top: 0',
    ],
    'code_inline': [
        'margin-left: 0.25em',
        'margin-right: 0.25em',
    ],
    'math_inline': [
        'margin-left: 0.25em',
        'margin-right: 0.25em',
    ],
}

class QuizComposerRendererHTML(markdown_it.renderer.RendererHTML):
    def render(self, tokens, options, env):
        # Override the main rendering function to attatch style.
        self._style_override_helper(tokens)
        return super().render(tokens, options, env)

    def _style_override_helper(self, tokens):
        if (tokens is None):
            return

        for token in tokens:
            if (token.type in ADDITIONAL_STYLE):
                _join_html_style(token, ADDITIONAL_STYLE.get(token.type, []))

            self._style_override_helper(token.children)

    def image(self, tokens, idx, options, env,
            force_raw_image_src = False, process_token = None):
        # Do custom rendering and then pass onto super.

        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})

        base_dir = context.get(quizcomp.parser.common.BASE_DIR_KEY, '.')
        callback = context.get(quizcomp.parser.common.CONTEXT_KEY_IMAGE_CALLBACK, None)

        # Set width.
        width_float = quizcomp.parser.style.get_image_width(style)
        tokens[idx].attrSet('width', "%0.2f%%" % (width_float * 100.0))

        original_src = tokens[idx].attrGet('src')
        src = quizcomp.parser.image.handle_callback(callback, original_src, base_dir)
        path = os.path.realpath(os.path.join(base_dir, src))
        tokens[idx].attrSet('src', src)

        # Check the env to see if we need to force raw images.
        force_raw_image_src = force_raw_image_src or context.get(quizcomp.parser.common.CONTEXT_KEY_FORCE_RAW_IMAGE_SRC, False)

        if (force_raw_image_src or re.match(r'^http(s)?://', src) or src.startswith('data:image')):
            # Do not further modify the src if we are explicitly directed not to
            # or if it is an http or data URL.
            pass
        else:
            # Otherwise, do a base64 encoding of the image and embed it.
            mime, content = quizcomp.parser.image.encode_image(path)
            tokens[idx].attrSet('src', f"data:{mime};base64,{content}")

        # Last chance to change the token before HTML rendering.
        if (process_token is not None):
            tokens[idx] = process_token(tokens[idx], context, path)

        result = super().image(tokens, idx, options, env)

        # Reset the src so that future callback hits have the proper cache key.
        tokens[idx].attrSet('src', original_src)

        return result

    def container_block_open(self, tokens, idx, options, env):
        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})

        # Add on a specific class.
        tokens[idx].attrJoin('class', 'qg-block')

        # Pull any style attatched to this block and put it in a copy of the context.
        context, full_style, block_style = quizcomp.parser.common.handle_block_style(tokens[idx].meta, context)
        env[quizcomp.parser.common.CONTEXT_ENV_KEY] = context

        # Attatch style based on if we are the root block.
        # If root use all style, otherwise just use the style for this block.
        active_style = block_style
        if (tokens[idx].meta.get(quizcomp.parser.common.TOKEN_META_KEY_ROOT, False)):
            active_style = full_style

        style_string = quizcomp.parser.style.compute_html_style_string(active_style)
        if (style_string != ''):
            tokens[idx].attrSet('style', style_string)

        # Send to super for further rendering.
        return super().renderToken(tokens, idx, options, env)

    def math_inline(self, tokens, idx, options, env):
        return quizcomp.parser.math.render(quizcomp.constants.FORMAT_HTML, True, tokens, idx, options, env)

    def math_block(self, tokens, idx, options, env):
        return quizcomp.parser.math.render(quizcomp.constants.FORMAT_HTML, False, tokens, idx, options, env)

    def placeholder(self, tokens, idx, options, env):
        return "<placeholder>%s</placeholder>" % (tokens[idx].content)

    def table_open(self, tokens, idx, options, env):
        token = tokens[idx]
        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})

        table_style = [
            'border-collapse: collapse',
        ]

        if (quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_BORDER_TABLE, quizcomp.parser.style.DEFAULT_TABLE_BORDER_TABLE)):
            table_style.append("border: %s" % HTML_BORDER_SPEC)
        else:
            table_style.append('border-style: hidden')

        # HTML tables require extra encouragement to align.
        text_align = quizcomp.parser.style.get_alignment(style, quizcomp.parser.style.KEY_TEXT_ALIGN)
        if (text_align is not None):
            table_style.append("text-align: %s" % (text_align))

        _join_html_style(token, table_style)

        return super().renderToken(tokens, idx, options, env)

    def thead_open(self, tokens, idx, options, env):
        token = tokens[idx]
        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})

        if (quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_HEAD_RULE, quizcomp.parser.style.DEFAULT_TABLE_HEAD_RULE)):
            _join_html_style(token, ["border-bottom: %s" % (HTML_BORDER_SPEC)])

        return super().renderToken(tokens, idx, options, env)

    def th_open(self, tokens, idx, options, env):
        token = tokens[idx]
        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})

        self._cell_html(token, style)

        weight = 'normal'
        if (quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_HEAD_BOLD, quizcomp.parser.style.DEFAULT_TABLE_HEAD_BOLD)):
            weight = 'bold'

        _join_html_style(token, ["font-weight: %s" % (weight)])

        return super().renderToken(tokens, idx, options, env)

    def td_open(self, tokens, idx, options, env):
        token = tokens[idx]
        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
        style = context.get(quizcomp.parser.common.CONTEXT_KEY_STYLE, {})

        self._cell_html(token, style)

        return super().renderToken(tokens, idx, options, env)

    def _cell_html(self, token, style):
        """
        Common cell rendering.
        """

        height = max(1.0, float(style.get(quizcomp.parser.style.KEY_TABLE_CELL_HEIGHT, quizcomp.parser.style.DEFAULT_TABLE_CELL_HEIGHT)))
        vertical_padding = height - 1.0

        width = max(1.0, float(style.get(quizcomp.parser.style.KEY_TABLE_CELL_WIDTH, quizcomp.parser.style.DEFAULT_TABLE_CELL_WIDTH)))
        horizontal_padding = width - 1.0

        cell_style = {
            'padding-top': "%0.2fem" % (vertical_padding / 2),
            'padding-bottom': "%0.2fem" % (vertical_padding / 2),
            'padding-left': "%0.2fem" % (horizontal_padding / 2),
            'padding-right': "%0.2fem" % (horizontal_padding / 2),
        }

        if (quizcomp.parser.style.get_boolean_style_key(style, quizcomp.parser.style.KEY_TABLE_BORDER_CELLS, quizcomp.parser.style.DEFAULT_TABLE_BORDER_CELLS)):
            cell_style['border'] = "%s" % (HTML_BORDER_SPEC)

        _join_html_style(token, [': '.join(item) for item in cell_style.items()])

def get_renderer(options):
    return QuizComposerRendererHTML(), options

def _join_html_style(token, rules):
    """
    Take all style rules to apply, add in any existing style, and set the style attribute.
    """

    if (len(rules) == 0):
        return

    existing_style = token.attrGet('style')
    if (existing_style is not None):
        rules = [existing_style] + rules

    style_string = '; '.join(rules)
    token.attrSet('style', style_string)



import os

import git

UNKNOWN_VERSION = 'UNKNOWN'
VERSION_LEN = 8
DIRTY_SIFFIX = '-d'

def get_version(path = '.', throw = False):
    """
    Get a version string from the git repo.
    This is just a commit hash with some dressup.
    """

    if (os.path.isfile(path)):
        path = os.path.dirname(path)

    try:
        repo = git.Repo(path, search_parent_directories = True)
    except Exception as ex:
        if (throw):
            raise ValueError(f"Path '{path}' is not a valid Git repo.") from ex

        return UNKNOWN_VERSION

    version = repo.head.commit.tree.hexsha[:VERSION_LEN]

    if (repo.is_dirty()):
        version += DIRTY_SIFFIX

    return version



import markdown_it

import quizcomp.parser.common

# Pull specific attributes from nodes of these specified types.
AST_NODE_ATTRIBUTES = {
    'code_block': [
        'info',
    ],
    'fence': [
        'info',
    ],
    'heading': [
        'tag',
    ],
}

# Pull these attributes out of these specific token types when building the AST.
AST_TOKEN_ATTRS = {
    'image': [
        'src',
    ],
    'link': [
        'href',
    ],
    'td': [
        'style',
    ],
    'th': [
        'style',
    ],
}

# Like AST_TOKEN_ATTRS, but for the `meta` property.
AST_TOKEN_METAS = {
    'container_block': [
        quizcomp.parser.common.TOKEN_META_KEY_ROOT,
        quizcomp.parser.common.TOKEN_META_KEY_STYLE,
    ]
}

class ASTNode(dict):
    def type(self):
        return self['type']

    def children(self):
        return self.get('children', [])

    def text(self):
        return self.get('text', '')

def build(tokens):
    tree = markdown_it.tree.SyntaxTreeNode(tokens)
    return _walk_ast(tree)

def _walk_ast(node):
    result = {
        'type': node.type,
    }

    if (node.type in quizcomp.parser.common.CONTENT_NODES):
        result['text'] = node.content

    for name in AST_NODE_ATTRIBUTES.get(node.type, []):
        value = getattr(node, name)
        if ((value is not None) and (value != '')):
            result[name] = value

    for name in AST_TOKEN_ATTRS.get(node.type, []):
        value = node.attrGet(name)
        if (value is not None):
            result[name] = value

    for name in AST_TOKEN_METAS.get(node.type, []):
        value = node.meta.get(name, None)
        if (value is not None):
            result[name] = value

    if (len(node.children) > 0):
        result['children'] = [_walk_ast(child) for child in node.children]

    return ASTNode(result)



import os
import sys

import quizcomp.args
import quizcomp.quiz
import quizcomp.uploader.canvas

DEFAULT_BASE_URL = 'https://canvas.ucsc.edu'

def run(args):
    if (not os.path.exists(args.path)):
        raise ValueError(f"Provided path '{args.path}' does not exist.")

    if (not os.path.isfile(args.path)):
        raise ValueError(f"Provided path '{args.path}' is not a file.")

    quiz = quizcomp.quiz.Quiz.from_path(args.path)
    canvas_instance = quizcomp.uploader.canvas.InstanceInfo(args.base_url, args.course_id, args.token)

    uploader = quizcomp.uploader.canvas.CanvasUploader(canvas_instance, force = args.force)
    uploader.upload_quiz(quiz)

    return 0

def _get_parser():
    parser = quizcomp.args.Parser(description =
        "Parse a quiz and upload the quiz to Canvas.")

    parser.add_argument('path', metavar = 'PATH',
        type = str,
        help = 'The path to a quiz json file.')

    parser.add_argument('--course', dest = 'course_id',
        action = 'store', type = str, required = True,
        help = 'Course ID to upload the quiz under.')

    parser.add_argument('--url', dest = 'base_url',
        action = 'store', type = str, default = DEFAULT_BASE_URL,
        help = 'The base URL for the Canvas instance (default: %(default)s).')

    parser.add_argument('--token', dest = 'token',
        action = 'store', type = str, required = True,
        help = 'The authentication token to use with Canvas.')

    parser.add_argument('--force', dest = 'force',
        action = 'store_true', default = False,
        help = 'Override (delete) any exiting quiz with the same name.')

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import abc
import copy
import importlib
import logging
import math
import os
import pkgutil
import random
import re

import quizcomp.common
import quizcomp.constants
import quizcomp.parser.public
import quizcomp.question.common
import quizcomp.util.dirent
import quizcomp.util.serial

BASE_MODULE_NAME = 'quizcomp.question'
THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

class Question(quizcomp.util.serial.JSONSerializer):
    # {question_type: class, ...}
    _types = {}
    _imported_this_package = False

    def __init_subclass__(cls, question_type = None, **kwargs):
        """
        Register question subclasses (types).
        """

        super().__init_subclass__(**kwargs)

        if (question_type is None):
            raise quizcomp.common.QuizValidationError("No question type provided for question subclass.")

        cls._types[question_type] = cls

    def __init__(self, type = quizcomp.constants.TYPE_QUESTION,
            prompt = '', prompt_path = None,
            question_type = '', answers = None,
            base_dir = '.',
            points = 0, name = '',
            shuffle_answers = True,
            custom_header = None, skip_numbering = None,
            hints = None, feedback = None,
            ids = {},
            **kwargs):
        super().__init__(type = type, **kwargs)

        self.base_dir = base_dir

        self.prompt = prompt
        self._prompt_path = prompt_path

        self.question_type = question_type

        self.answers = answers

        self.points = points
        self.name = name

        self.hints = hints
        self.feedback = feedback

        self.shuffle_answers = shuffle_answers

        self.custom_header = custom_header
        self.skip_numbering = skip_numbering

        self.ids = ids.copy()
        self.ids['base_dir'] = base_dir

        try:
            self.validate()
        except Exception as ex:
            ids = self.ids.copy()
            ids['name'] = self.name
            ids['question_type'] = self.question_type

            raise quizcomp.common.QuizValidationError('Error while validating question.', ids = ids) from ex

    def _validate(self):
        self._validate_prompt()
        self._validate_question_feedback()
        self._validate_answers()

        if (self.hints is None):
            self.hints = {}
        else:
            self._check_type(self.hints, dict, "'hints'")

    @abc.abstractmethod
    def _validate_answers(self):
        pass

    def _validate_prompt(self):
        """
        The prompt is allowed to appear (in order of priority):
        in the prompt field, be pointed to by the _prompt_path member, or be in ./quizcomp.constants.PROMPT_FILENAME.
        Both None and empty/white strings are ignored.

        Will raise an exception on an empty prompt.
        If this method does not raise an exception, the result will be placed in self.prompt.
        If the prompt is loaded from a file, self._prompt_path will be set with the absolute path.
        """

        text = self._get_prompt_text()
        self.prompt = self._validate_text_item(text, 'question prompt', check_feedback = False, allow_empty = False)

    def _get_prompt_text(self):
        # First check self.prompt.
        text = self.prompt
        if (text is None):
            text = ''

        text = text.strip()

        if (text != ''):
            return text

        # Next, check for prompt files.

        check_paths = [self._prompt_path, quizcomp.constants.PROMPT_FILENAME]
        for path in check_paths:
            if ((path is None) or (path.strip() == '')):
                continue

            if (not os.path.isabs(path)):
                path = os.path.join(self.base_dir, path)

            path = os.path.abspath(path)
            if (not os.path.exists(path)):
                continue

            logging.debug("Loading question prompt from '%s'.", path)
            self._prompt_path = path
            return quizcomp.util.dirent.read_file(path)

        raise quizcomp.common.QuestionValidationError(self, "Could not find any non-empty prompt.")

    def inherit_from_group(self, group):
        """
        Inherit attributes from a group.
        """

        self.points = group.points
        self.name = group.name

        if (group.custom_header is not None):
            self.custom_header = group.custom_header

        if (group.skip_numbering is not None):
            self.skip_numbering = group.skip_numbering

        self.shuffle_answers = (self.shuffle_answers and group.shuffle_answers)

        self.add_hints(group.hints)

    def add_hints(self, new_hints, override = False):
        for (key, value) in new_hints.items():
            if (override or (key not in self.hints)):
                self.hints[key] = value

    def collect_file_paths(self):
        paths = []

        for document in self._collect_documents([self.prompt, self.answers]):
            paths += document.collect_file_paths(self.base_dir)

        return paths

    def _collect_documents(self, target):
        if (isinstance(target, dict)):
            return self._collect_documents(list(target.values()))
        elif (isinstance(target, list)):
            documents = []
            for value in target:
                documents += self._collect_documents(value)
            return documents
        elif (isinstance(target, quizcomp.parser.public.ParsedText)):
            return [target.document]
        else:
            return []

    def should_skip_numbering(self):
        return ((self.skip_numbering is not None) and (self.skip_numbering))

    def copy(self):
        return copy.deepcopy(self)

    def shuffle(self, rng = None):
        if (not self.shuffle_answers):
            return

        if (rng is None):
            rng = random.Random()

        self._shuffle(rng)

    def _shuffle(self, rng):
        """
        Shuffle the answers for this question.
        By default (this method), no shuffling is performed.
        Children can override this method to support shuffling.
        """

        pass

    def _shuffle_answers_list(self, rng):
        """
        A shuffle method for question types that are a simple list.
        """

        rng.shuffle(self.answers)

    # Override the class method JSONSerializer.from_dict() with a static method
    # so that we can select the correct child class.
    @staticmethod
    def from_dict(data, base_dir = None, ids = {}, **kwargs):
        if (base_dir is not None):
            data['base_dir'] = base_dir
        elif ('base_dir' not in data):
            data['base_dir'] = '.'

        if ('question_type' not in data):
            raise quizcomp.common.QuizValidationError("Question does not contain a 'question_type' field.", ids = ids)

        question_class = Question._fetch_question_class(data.get('question_type'), ids = ids, **kwargs)
        return quizcomp.util.serial._from_dict(question_class, data, ids = ids, **kwargs)

    @staticmethod
    def _fetch_question_class(question_type, ids = {}, **kwargs):
        if (not Question._imported_this_package):
            for _, name, is_package in pkgutil.iter_modules([THIS_DIR]):
                if (is_package):
                    continue

                module_name = BASE_MODULE_NAME + '.' + name
                importlib.import_module(module_name)

            _imported_this_package = True

        if (question_type not in Question._types):
            ids = ids.copy()
            ids['question_type'] = question_type

            raise quizcomp.common.QuizValidationError("Unknown question type.", ids = ids)

        return Question._types[question_type]

    def _validate_question_feedback(self):
        if (self.feedback is None):
            self.feedback = {}
            return

        if (isinstance(self.feedback, str)):
            self.feedback = {'general': self.feedback}

        self._check_type(self.feedback, dict, "'feedback'")

        allowed_keys = ['general', 'correct', 'incorrect']
        actual_keys = list(self.feedback.keys())

        bad_keys = list(sorted(set(actual_keys) - set(allowed_keys)))
        if (len(bad_keys) > 0):
            raise quizcomp.common.QuestionValidationError(self, "Unknown keys in feedback (%s). Allowed keys: '%s'." % (
                    str(bad_keys), str(allowed_keys)))

        new_feedback = {}
        for (key, value) in self.feedback.items():
            item = self._validate_feedback_item(value, "'%s' feedback value" % (key))
            if (item is not None):
                new_feedback[key] = item

        self.feedback = new_feedback

    def _validate_feedback_item(self, item, label):
        if ((item is None) or isinstance(item, quizcomp.parser.public.ParsedText)):
            # Nothing to do.
            return item

        self._check_type(item, str, label)

        item = item.strip()
        if (len(item) == 0):
            return None

        return quizcomp.parser.public.parse_text(item, base_dir = self.base_dir)

    def _validate_self_answer_list(self, min_correct = 0, max_correct = math.inf):
        self.answers = self._validate_answer_list(self.answers, self.base_dir,
                min_correct = min_correct, max_correct = max_correct)

    def _validate_answer_list(self, answers, base_dir, min_correct = 0, max_correct = math.inf):
        self._check_type(answers, list, "'answers'")

        if (len(answers) == 0):
            raise quizcomp.common.QuestionValidationError(self, f"No answers provided, at least one answer required.")

        num_correct = 0
        for answer in answers:
            if ('correct' not in answer):
                raise quizcomp.common.QuestionValidationError(self, f"Answer has no 'correct' field.")

            if ('text' not in answer):
                raise quizcomp.common.QuestionValidationError(self, f"Answer has no 'text' field.")

            if (answer['correct']):
                num_correct += 1

        if (num_correct < min_correct):
            raise quizcomp.common.QuestionValidationError(self, f"Did not find enough correct answers. Expected at least {min_correct}, found {num_correct}.")

        if (num_correct > max_correct):
            raise quizcomp.common.QuestionValidationError(self, f"Found too many correct answers. Expected at most {max_correct}, found {num_correct}.")

        new_answers = []
        for i in range(len(answers)):
            parsed_text = self._validate_text_item(answers[i], "'answers' values (element %d)" % (i))
            new_answer = quizcomp.question.common.ParsedTextChoice(parsed_text, answers[i]['correct'])
            new_answers.append(new_answer)

        return new_answers

    def _validate_text_answers(self):
        possible_answers = 'null/None, string, empty list, list of strings, or list of objects'

        if (self.answers is None):
            self.answers = ['']
        elif (isinstance(self.answers, str)):
            self.answers = [self.answers]

        if (not isinstance(self.answers, list)):
            raise quizcomp.common.QuestionValidationError(self, "'answers' value must be %s, found: '%s'." % (
                    possible_answers, str(self.answers)))

        if (len(self.answers) == 0):
            self.answers = ['']

        new_answers = []
        for i in range(len(self.answers)):
            new_answers.append(self._validate_text_item(self.answers[i], "'answers' values (element %d)" % (i)))

        self.answers = new_answers

    def _validate_text_item(self, item, label,
            check_feedback = True, allow_empty = True,
            strip = True, clean_whitespace = False):
        """
        Validate a portion of an answer/choice/field that is a parsed string.

        Allowed values are:
         - None (will be converted to an empty string).
         - Empty String (if allow_empty is True).
         - String
         - quizcomp.question.common.ParsedTextWithFeedback (will be passed back without any checks).
         - Dict with required key 'text' and optional key 'feedback'.

        If no exception is raised, a quizcomp.question.common.ParsedTextWithFeedback (child of quizcomp.parser.public.ParsedText)
        will be returned, even if there is no feedback.
        """

        if (isinstance(item, quizcomp.question.common.ParsedTextWithFeedback)):
            # Nothing to do if the item is already parsed.
            return item

        if (item is None):
            item = ''

        if (isinstance(item, str)):
            item = {'text': item}

        self._check_type(item, dict, label)

        if ('text' not in item):
            raise quizcomp.common.QuestionValidationError(self, "%s is missing a 'text' key." % (label))

        text = item['text']
        self._check_type(item['text'], str, "%s 'text' key" % (label))

        if (clean_whitespace):
            text = re.sub(r'\s+', ' ', text)

        if (strip):
            text = text.strip()

        if ((not allow_empty) and (text == '')):
            raise quizcomp.common.QuestionValidationError(self, "%s text is empty." % (label))

        feedback = None
        if (check_feedback):
            feedback = self._validate_feedback_item(item.get('feedback', None), label)

        return quizcomp.question.common.ParsedTextWithFeedback(quizcomp.parser.public.parse_text(text, base_dir = self.base_dir), feedback = feedback)

    def _validate_fimb_answers(self):
        self._check_type(self.answers, dict, "'answers' key")

        if (len(self.answers) == 0):
            raise quizcomp.common.QuestionValidationError(self, "Expected 'answers' dict to be non-empty.")

        for (key, values) in self.answers.items():
            # If this was already in the full FIMB format, then we need to pull out the values.
            if ((isinstance(values, dict)) and ('values' in values)):
                self.answers[key] = values['values']
            elif (not isinstance(values, list)):
                self.answers[key] = [values]

        new_answers = {}

        for (key, values) in self.answers.items():
            self._check_type(key, str, "key in 'answers' dict")

            if (len(values) == 0):
                raise quizcomp.common.QuestionValidationError(self, "Expected possible values to be non-empty.")

            new_values = []
            for i in range(len(values)):
                label = "answers key '%s' index %d" % (key, i)
                new_values.append(self._validate_text_item(values[i], label))

            new_answers[key] = {
                'key': quizcomp.parser.public.parse_text(key, base_dir = self.base_dir),
                'values': new_values,
            }

        self.answers = new_answers

        self._check_placeholders(self.answers.keys())

    def _check_type(self, value, expected_type, label):
        if (not isinstance(value, expected_type)):
            raise quizcomp.common.QuestionValidationError(self, f"{label} must be a {expected_type}, found '{value}' ({type(value)}).")

    def _check_placeholders(self, answer_placeholders):
        """
        Check placeholders from the answers against placeholders in the prompt.
        """

        answer_placeholders = set(list(answer_placeholders))
        document_placeholders = self.prompt.document.collect_placeholders()

        # Special case for FITB documents.
        if ((len(answer_placeholders) == 1) and (list(answer_placeholders)[0] == '')):
            if (len(document_placeholders) != 0):
                output_answer_placeholders = list(sorted(answer_placeholders))
                raise quizcomp.common.QuestionValidationError(self, "Found placeholders (%s) in the question prompt when none were expected." % (output_answer_placeholders))

            return

        if (answer_placeholders != document_placeholders):
            output_answer_placeholders = list(sorted(answer_placeholders))
            output_document_placeholders = list(sorted(document_placeholders))

            raise quizcomp.common.QuestionValidationError(self, "Mismatch between the placeholders found in the question prompt (%s) and answers config (%s)." % (output_document_placeholders, output_answer_placeholders))



import re

import quizcomp.parser.common
import quizcomp.parser.image
import quizcomp.parser.renderer.html

class QuizComposerRendererCanvas(quizcomp.parser.renderer.html.QuizComposerRendererHTML):
    """
    Canvas generally uses HTML, but has some special cases.
    """

    def placeholder(self, tokens, idx, options, env):
        # Canvas placeholders cannot have spaces.
        text = tokens[idx].content.strip()
        text = re.sub(r'\s+', ' ', text)
        text = text.replace(' ', '_')

        return "[%s]" % (text)

    def image(self, tokens, idx, options, env):
        # Canvas requires files to be uploaded instead of embedded.
        # Those files should have already been uploaded and available.

        context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})

        force_raw_image_src = True
        process_token = _process_token

        # If there is no canvas instance, we are probably just parsing and not uploading.
        canvas_instance = context.get('canvas_instance', None)
        if (canvas_instance is None):
            force_raw_image_src = False
            process_token = None

        return super().image(tokens, idx, options, env,
                force_raw_image_src = force_raw_image_src,
                process_token = process_token)

def _process_token(token, context, path):
    canvas_instance = context.get('canvas_instance', None)
    if (canvas_instance is None):
        raise ValueError('Could not get canvas context.')

    file_id = canvas_instance.context.get('file_ids', {}).get(path)
    if (file_id is None):
        raise ValueError(f"Could not get canvas context file id of image '{path}'.")

    token.attrSet('src', f"{canvas_instance.base_url}/courses/{canvas_instance.course_id}/files/{file_id}/preview")
    return token

def get_renderer(options):
    return QuizComposerRendererCanvas(), options



import abc
import copy as pycopy
import datetime
import os

import quizcomp.util.dirent
import quizcomp.util.json

class PODSerializer(abc.ABC):
    @abc.abstractmethod
    def to_pod(self, **kwargs):
        """
        Create a "Plain Old Data" representation of this object.
        """

        pass

class JSONSerializer(PODSerializer):
    """
    A base class that can automatically handle serialization.
    Deserialization is harder (and requires validation),
    so will be left as abstract.
    """

    def __init__(self, type = 'unknown', _skip_all_validation = False, _skip_class_validations = [], **kwargs):
        # A marked type that can be useful for deserialization.
        self.type = type

        self._skip_all_validation = _skip_all_validation

        # Keep track of the classes that have been validated, so they can be skipped.
        self._validated_classes = {}

        for cls in _skip_class_validations:
            self._validated_classes[cls] = True

    def validate(self, cls = None, **kwargs):
        """
        A wrapper for validation.
        This should be called by child classes in their constructor.
        If cls is provided, then that specific _validate will be called.
        Otherwise, whatever default _validate() is registered for self's class will be called.
        This method should raise an exception if the object is invalid,
        and set self._validated_classes[cls] = True if the object is valid.
        """

        if (self._skip_all_validation):
            return

        if (cls is None):
            cls = self.__class__

        if (self._validated_classes.get(cls, False)):
            return

        cls._validate(self, **kwargs)
        self._validated_classes[cls] = True

    @abc.abstractmethod
    def _validate(self, **kwargs):
        """
        The true validation implementation.
        """

        pass

    def to_pod(self, **kwargs):
        return self.to_dict(**kwargs)

    def to_dict(self, copy = True, **kwargs):
        """
        Convert self to a dictonary that can easily be serialized.
        See _serialize() for all the keyword arguments.
        """

        data = self.__dict__

        if (copy):
            data = pycopy.deepcopy(data)

        return _serialize(data, **kwargs)

    def to_json(self, indent = 4, sort_keys = True, **kwargs):
        data = self.to_dict(**kwargs)
        return quizcomp.util.json.dumps(data, indent = indent, sort_keys = sort_keys)

    def to_path(self, path, **kwargs):
        quizcomp.util.dirent.write_file(path, self.to_json(**kwargs))

    @classmethod
    def from_dict(cls, data, copy = True, extra_fields = {}, **kwargs):
        return _from_dict(cls, data, copy = copy, extra_fields = extra_fields, **kwargs)

    @classmethod
    def from_path(cls, path, add_base_dir = True, data_callback = None, **kwargs):
        path = os.path.abspath(path)
        ids = {
            'path': path,
        }

        if (not os.path.isfile(path)):
            raise quizcomp.common.QuizValidationError('Path does not exist or is not a file.', ids = ids)

        try:
            data = quizcomp.util.json.load_path(path)
        except Exception as ex:
            raise quizcomp.common.QuizValidationError('Failed to read JSON file (invalid JSON?).', ids = ids) from ex

        base_dir = os.path.dirname(os.path.abspath(path))
        if (('base_dir' not in data) and add_base_dir):
            data['base_dir'] = base_dir

        if (data_callback is not None):
            data = data_callback(path, data)

        return cls.from_dict(data, copy = False, base_dir = base_dir, ids = ids, **kwargs)

def _from_dict(cls, data, copy = True, extra_fields = {}, **kwargs):
    if (copy):
        data = pycopy.deepcopy(data)

    for (key, value) in extra_fields.items():
        data[key] = value

    return cls(**data)

def _serialize(item,
        skip_private = True,
        convert_serializers = True,
        convert_dates = True,
        recursive = True,
        **kwargs):
    """
    Generally convert an object into a form better suited to serialization.
    """

    kwargs['skip_private'] = skip_private
    kwargs['convert_serializers'] = convert_serializers
    kwargs['convert_dates'] = convert_dates
    kwargs['recursive'] = recursive

    if (isinstance(item, PODSerializer) and convert_serializers):
        return item.to_pod(**kwargs)
    elif (isinstance(item, list) and recursive):
        return [_serialize(value, **kwargs) for value in item]
    elif (isinstance(item, dict) and recursive):
        return {key: _serialize(value, **kwargs) for (key, value) in item.items() if (not (skip_private and key.startswith('_')))}
    elif (isinstance(item, (datetime.date, datetime.time, datetime.datetime)) and convert_dates):
        return item.isoformat()

    return item



import base64
import os

import quizcomp.parser.common

def handle_callback(callback, original_src, base_dir):
    if (callback is None):
        return original_src

    return callback(original_src, base_dir)

def encode_image(path):
    ext = os.path.splitext(path)[-1].lower()
    ext = ext.removeprefix('.')
    mime = f"image/{ext}"

    with open(path, 'rb') as file:
        data = file.read()

    content = base64.standard_b64encode(data)
    return mime, content.decode(quizcomp.parser.common.ENCODING)



"""
Convert a quiz into HTML using templates.
"""

import os

import quizcomp.constants
import quizcomp.converter.template

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DEFAULT_TEMPLATE_DIR = os.path.join(THIS_DIR, '..', 'data', 'templates', 'edq-html')

class HTMLTemplateConverter(quizcomp.converter.template.TemplateConverter):
    def __init__(self,
            format = quizcomp.constants.FORMAT_HTML, template_dir = DEFAULT_TEMPLATE_DIR,
            **kwargs):
        super().__init__(format, template_dir, **kwargs)

    def clean_solution_content(self, document):
        return document.to_text()

class CanvasTemplateConverter(HTMLTemplateConverter):
    def __init__(self,
            template_dir = DEFAULT_TEMPLATE_DIR,
            **kwargs):
        super().__init__(quizcomp.constants.FORMAT_CANVAS, template_dir, **kwargs)



import quizcomp.common
import quizcomp.constants
import quizcomp.question.base

class TF(quizcomp.question.base.Question, question_type = quizcomp.constants.QUESTION_TYPE_TF):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _validate_answers(self):
        if (isinstance(self.answers, bool)):
            # Change answers to look like multiple choice.
            self.answers = [
                {"correct": self.answers, "text": 'True'},
                {"correct": (not self.answers), "text": 'False'},
            ]
        elif (isinstance(self.answers, list)):
            pass
        else:
            raise quizcomp.common.QuestionValidationError(self, f"'answers' value must be a boolean, found '{self.answers}' ({type(self.answers)}).")

        self._validate_self_answer_list()

        if (len(self.answers) != 2):
            raise quizcomp.common.QuestionValidationError(self, "Expecting exactly two answers, found %d." % (len(self.answers)))

        labels = list(sorted([answer.text for answer in self.answers]))

        expected = ['False', 'True']
        if (labels != expected):
            raise quizcomp.common.QuestionValidationError(self, "T/F labels (text) not as expected. Expected: '%s', Actual: '%s'." % (expected, labels))

    def _shuffle(self, rng):
        self._shuffle_answers_list(rng)



import quizcomp.util.json

class QuizValidationError(ValueError):
    def __init__(self, message, ids = {}, **kwargs):
        ids = ids.copy()
        ids.update(kwargs)

        parsed_ids = {}
        for (key, value) in ids.items():
            if ((value is None) or (value == '')):
                continue

            parsed_ids[str(key)] = value

        if (len(parsed_ids) > 0):
            message = "%s %s" % (message, quizcomp.util.json.dumps(parsed_ids))

        super().__init__(message)

class QuestionValidationError(QuizValidationError):
    def __init__(self, question, message, **kwargs):
        super().__init__(message, ids = question.ids, **kwargs)



"""
Create a PDF quiz.
"""

import sys

import quizcomp.args
import quizcomp.pdf

def run(args):
    quizcomp.pdf.make_with_args(args)
    return 0

def _get_parser():
    parser = quizcomp.args.Parser(description =
        "Create a PDF quiz.")

    quizcomp.pdf.set_cli_args(parser)

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import re

import quizcomp.parser.common
import quizcomp.parser.renderer.base

DISALLOWED_CHARACTERS = re.compile(r'[^\w \-]')

class QuizComposerRendererText(quizcomp.parser.renderer.base.QuizComposerRendererBase):
    """
    The text renderer tries to output plan text that will then be used for special purposes like keys and identifiers.
    The output here is not meant to represent full documents or be sent to users.
    """

    def clean_final(self, text, context):
        allow_all_characters = context.get(quizcomp.parser.common.CONTEXT_KEY_TEXT_ALLOW_ALL_CHARACTERS, False)

        # Clean up whitespace.
        text = re.sub(r'\s+', ' ', text).strip()

        # Remove bad disallowed characters.
        if (not allow_all_characters):
            text = re.sub(DISALLOWED_CHARACTERS, '', text)

            # Clean up whitespace once more.
            text = re.sub(r'\s+', ' ', text).strip()

        return text

    def _text(self, node, context):
        return _clean_text(node.text())

    def _softbreak(self, node, context):
        return "\n"

    def _hardbreak(self, node, context):
        return "\n\n"

    def _em(self, node, context):
        return ''.join([self._render_node(child, context) for child in node.children()])

    def _strong(self, node, context):
        return ''.join([self._render_node(child, context) for child in node.children()])

    def _fence(self, node, context):
        return self._handle_special_text(node, context)

    def _code_block(self, node, context):
        return self._handle_special_text(node, context)

    def _code_inline(self, node, context):
        return self._handle_special_text(node, context)

    def _math_block(self, node, context):
        return self._handle_special_text(node, context)

    def _math_inline(self, node, context):
        return self._handle_special_text(node, context)

    def _image(self, node, context):
        return ''

    def _link(self, node, context):
        return ''

    def _placeholder(self, node, context):
        return _clean_text(node.text())

    def _table(self, node, context):
        return ''

    def _thead(self, node, context):
        return ''

    def _tbody(self, node, context):
        return ''

    def _tr(self, node, context):
        return ''

    def _th(self, node, context):
        return ''

    def _td(self, node, context):
        return ''

    def _bullet_list(self, node, context):
        return ''

    def _ordered_list(self, node, context):
        return ''

    def _list_item(self, node, context):
        return ''

    def _hr(self, node, context):
        return ''

    def _heading(self, node, context):
        return ''.join([self._render_node(child, context) for child in node.children()])

    def _blockquote(self, node, context):
        return ''.join([self._render_node(child, context) for child in node.children()])

    def _handle_special_text(self, node, context):
        """
        Special text is usually not allowed,
        but can be enabled with quizcomp.parser.common.CONTEXT_KEY_TEXT_ALLOW_SPECIAL_TEXT.
        """

        if (context.get(quizcomp.parser.common.CONTEXT_KEY_TEXT_ALLOW_SPECIAL_TEXT, False)):
            return node.text().strip()

        return ''

def get_renderer(options):
    return QuizComposerRendererText(), options

def _clean_text(text):
    return text



import os
import random
import sys

import quizcomp.args
import quizcomp.converter.convert
import quizcomp.quiz

def run(args):
    if (not os.path.exists(args.path)):
        raise ValueError(f"Provided path '{args.path}' does not exist.")

    if (not os.path.isfile(args.path)):
        raise ValueError(f"Provided path '{args.path}' is not a file.")

    for quiz_format in args.formats:
        if (quiz_format not in quizcomp.converter.convert.SUPPORTED_FORMATS):
            raise ValueError("Unknown quiz format '%s', must be one of: [%s]." % (quiz_format, ', '.join(quizcomp.converter.convert.SUPPORTED_FORMATS)))

    seed = args.seed
    if (seed is None):
        seed = random.randint(0, 2**64)

    print("Parsing quiz: '%s'." % (args.path))

    quiz = quizcomp.quiz.Quiz.from_path(args.path, flatten_groups = args.flatten_groups)
    variant = quiz.create_variant(all_questions = args.flatten_groups, seed = seed)

    for quiz_format in args.formats:
        print("Generating quiz content for '%s'." % (quiz_format))
        quizcomp.converter.convert.convert_variant(variant, format = quiz_format,
                constructor_args = {'answer_key': args.answer_key})

    return 0

def _get_parser():
    parser = quizcomp.args.Parser(description = ("Parse a quiz for the purposes of testing."
            + " The quiz will be parsed one and content will be generated for (but not output)"
            + " for each of the specified format (or none if none are specified)."))

    parser.add_argument('path', metavar = 'PATH',
        type = str,
        help = 'The path to a quiz json file.')

    parser.add_argument('formats', metavar = 'FORMAT',
        type = str, nargs = '*',
        help = 'Generate (but do not output) content in this format.')

    parser.add_argument('--key', dest = 'answer_key',
        action = 'store_true', default = False,
        help = 'Generate an answer key instead of a blank quiz (default: %(default)s).')

    parser.add_argument('--flatten-groups', dest = 'flatten_groups',
        action = 'store_true', default = False,
        help = 'Flatten question groups with multiple questions to multiple groups with a single question (default: %(default)s).')

    parser.add_argument('--seed', dest = 'seed',
        action = 'store', type = int, default = None,
        help = 'The random seed to use (defaults to a random seed).')

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import os

import quizcomp.cli.gradescope.upload
import quizcomp.latex
import quizcomp.uploader.gradescope
import quizcomp.util.httpsession
import tests.base

SIMPLE_SESSION_NAME = 'gradescope-simple'
ALREADY_EXISTS_SESSION_NAME = 'gradescope-already-exists'
VARIANT_FORCE_RUBRIC_A_SESSION_NAME = 'gradescope-variant-force-rubric-a'
VARIANT_FORCE_RUBRIC_B_SESSION_NAME = 'gradescope-variant-force-rubric-b'
VARIANT_FORCE_RUBRIC_GROUP_SESSION_NAME = 'gradescope-variant-force-rubric-group'

TEST_COURSE = '100001'
TEST_USER = 'test@test.edulinq.org'
TEST_PASS = 'abc123'

class TestUploaderGradescope(tests.base.BaseTest):
    """
    Test uploading GradeScope quizzes using recorded HTTP sessions.
    """

    def _run_session_upload_test(self, session_infos, quiz_path, extra_args = []):
        if (not quizcomp.latex.is_available()):
            self.skipTest("LaTeX is not available.")

        for (session_id, session_dirname) in session_infos:
            session_base_dir = os.path.join(tests.base.HTTP_SESSIONS_DIR, session_dirname)

            quizcomp.util.httpsession.load_test_session(session_id, session_base_dir)

        temp_dir = quizcomp.util.dirent.get_temp_path('quizcomp-test-')

        raw_args = [
            '--log-level', 'ERROR',
            '--course', TEST_COURSE,
            '--user', TEST_USER,
            '--pass', TEST_PASS,
            '--out', temp_dir,
            quiz_path,
        ]

        raw_args += extra_args

        parser = quizcomp.cli.gradescope.upload._get_parser()
        args = parser.parse_args(raw_args)

        quizcomp.cli.gradescope.upload.run(args)

    def test_simple(self):
        session_infos = [
            (quizcomp.uploader.gradescope.SESSION_ID_UPLOAD, SIMPLE_SESSION_NAME),
        ]

        quiz_path = os.path.abspath(os.path.join(tests.base.GOOD_QUIZZES_DIR, 'single-question', 'quiz.json'))

        self._run_session_upload_test(session_infos, quiz_path)

    def test_already_exists(self):
        session_infos = [
            (quizcomp.uploader.gradescope.SESSION_ID_UPLOAD, ALREADY_EXISTS_SESSION_NAME),
        ]

        quiz_path = os.path.abspath(os.path.join(tests.base.GOOD_QUIZZES_DIR, 'single-question', 'quiz.json'))

        self._run_session_upload_test(session_infos, quiz_path)

    def test_variant_force_rubric(self):
        session_infos = [
            (quizcomp.uploader.gradescope.SESSION_ID_UPLOAD, VARIANT_FORCE_RUBRIC_A_SESSION_NAME),
            (quizcomp.uploader.gradescope.SESSION_ID_UPLOAD, VARIANT_FORCE_RUBRIC_B_SESSION_NAME),
            (quizcomp.uploader.gradescope.SESSION_ID_CREATE_ASSIGNMENT_GROUP, VARIANT_FORCE_RUBRIC_GROUP_SESSION_NAME),
        ]

        quiz_path = os.path.abspath(os.path.join(tests.base.GOOD_QUIZZES_DIR, 'single-question', 'quiz.json'))

        extra_args = [
            '--variants', '2',
            '--force',
            '--rubric',
        ]

        self._run_session_upload_test(session_infos, quiz_path, extra_args = extra_args)



import os
import re

import quizcomp.constants
import quizcomp.parser.common
import quizcomp.parser.public
import quizcomp.util.json
import tests.base

SKIP_COMMONMARK_TESTS = {
    176,  # Has non-JSON HTML style.
}

class TestParser(tests.base.BaseTest):
    """
    Test parsing.
    Good and bad situations will be loaded below into individual test cases.
    """

    pass

def _add_good_parse_questions():
    for path in tests.base.discover_good_document_files():
        with open(path, 'r') as file:
            documents = quizcomp.util.json.load(file)

        base_dir = os.path.dirname(path)

        for document in documents:
            name = document['name']
            text = document['text']

            for (doc_format, expected) in document['formats'].items():
                test_name = _make_name('good_parse', path, name, doc_format)
                options = document.get('options', {}).get(doc_format, {})
                context = document.get('context', {})
                setattr(TestParser, test_name, _get_good_parse_test(text, doc_format, expected, base_dir, options, context))

def _get_good_parse_test(text, doc_format, base_expected, base_dir, options, context):
    def __method(self):
        document = quizcomp.parser.public.parse_text(text).document
        result = document.to_format(doc_format, base_dir = base_dir, include_metadata = False, **context)

        if (doc_format == quizcomp.constants.FORMAT_JSON):
            result = quizcomp.util.json.loads(result)
            expected = {
                'type': 'document',
                'ast': {
                    'type': 'root',
                },
            }

            expected_children = []
            if (len(base_expected) > 0):
                if (isinstance(base_expected, list)):
                    expected_children += base_expected
                elif (isinstance(base_expected, dict)):
                    expected_children.append(base_expected)
                else:
                    raise ValueError("Unkown type for expected children: '%s'.", type(base_expected))

            if (len(expected_children) > 0):
                # If the first node is not the root block, then automatically insert it.
                if (not expected_children[0].get(quizcomp.parser.common.TOKEN_META_KEY_ROOT, False)):
                    expected_children = [{
                        'type': 'container_block',
                        quizcomp.parser.common.TOKEN_META_KEY_ROOT: True,
                        'children': expected_children,
                    }]

                expected['ast']['children'] = expected_children

            self.assertJSONDictEqual(expected, result)
        elif (doc_format in {quizcomp.constants.FORMAT_CANVAS, quizcomp.constants.FORMAT_HTML}):
            # If the HTML does not have a root block, then add one.
            raw_expected = base_expected
            if (options.get('strip', True)):
                raw_expected = raw_expected.strip()

            if ((raw_expected != '') and ('qg-root-block' not in raw_expected)):
                raw_expected = '<div class="qg-root-block qg-block">' + raw_expected + '</div>'

            expected = quizcomp.parser.render.clean_html(raw_expected, pretty = options.get('pretty', True))
            result = quizcomp.parser.render.clean_html(result, pretty = options.get('pretty', True))

            expected, result = _apply_text_options(options, expected, result)
            self.assertLongStringEqual(expected, result)
        else:
            expected = base_expected.strip()
            result = result.strip()

            expected, result = _apply_text_options(options, expected, result)
            self.assertEqual(expected, result)

    return __method

def _add_bad_parse_questions():
    for path in tests.base.discover_bad_document_files():
        with open(path, 'r') as file:
            documents = quizcomp.util.json.load(file)

        base_dir = os.path.dirname(path)

        for document in documents:
            name = document['name']
            text = document['text']
            options = document.get('options', {})

            test_name = _make_name('bad_parse', path, name)
            setattr(TestParser, test_name, _get_bad_parse_test(text, base_dir, options))

def _get_bad_parse_test(text, base_dir, options):
    def __method(self):
        try:
            quizcomp.parser.public.parse_text(text)
        except Exception:
            # Expected.
            return

        self.fail("Failed to raise an exception.")

    return __method

def _add_commonmark_tests():
    """
    Add test cases that come directly from the CommonMark spec.
    We won't try to validate the output, we just want to make sure they parse and render cleanly.
    The main thing with these tests is ensuring that our custom rendering does not fail.
    """

    with open(tests.base.COMMONMARK_TEST_DATA_PATH, 'r') as file:
        test_data = quizcomp.util.json.load(file)

    for test_case in test_data:
        id = test_case['example']

        if (id in SKIP_COMMONMARK_TESTS):
            continue

        text = test_case['markdown']
        section = _clean_name_part(test_case['section'])

        for format in quizcomp.constants.PARSER_FORMATS:
            name = "test_commonmark__%04d__%s__%s" % (id, section, format)
            setattr(TestParser, name, _get_commonmark_test(text, format))

def _get_commonmark_test(text, format):
    def __method(self):
        parsed_text = quizcomp.parser.public.parse_text(text)

        options = {
            # The examples use paths that we would try and encode.
            'force_raw_image_src': True,
        }
        parsed_text.document.to_format(format, **options)

    return __method

def _make_name(prefix, path, name, doc_format = None):
    clean_name = _clean_name_part(name)

    filename = os.path.splitext(os.path.basename(path))[0]

    test_name = "test_%s__%s__%s" % (prefix, filename, clean_name)

    if (doc_format is not None):
        test_name += ('__' + doc_format)

    return test_name

def _clean_name_part(text):
    clean_text = text.lower().strip().replace(' ', '_')
    clean_text = re.sub(r'\W+', '', clean_text)
    return clean_text

def _apply_text_options(options, a, b):
    if (options.get("ignore-whitespace", False)):
        a = re.sub(r'\s+', '', a)
        b = re.sub(r'\s+', '', b)

    return a, b

_add_bad_parse_questions()
_add_good_parse_questions()
_add_commonmark_tests()



import html

import quizcomp.constants
import quizcomp.katex
import quizcomp.parser.common

_katex_available = None

def render(format, inline, tokens, idx, options, env):
    context = env.get(quizcomp.parser.common.CONTEXT_ENV_KEY, {})
    text = tokens[idx].content

    if (format == quizcomp.constants.FORMAT_HTML):
        return _render_html(text, inline, context)
    elif (format == quizcomp.constants.FORMAT_MD):
        return _render_md(text, inline, context)
    elif (format == quizcomp.constants.FORMAT_TEX):
        return _render_tex(text, inline, context)
    elif (format == quizcomp.constants.FORMAT_TEXT):
        return _render_md(text, inline, context)
    else:
        raise ValueError(f"Unknown format '{format}'.")

def _render_tex(text, inline, context):
    text = text.replace('$', r'\$')

    if (inline):
        text = text.strip()
        return f"$ {text} $"

    return f"$$\n{text}\n$$"

def _render_html(text, inline, context):
    global _katex_available

    if (_katex_available is None):
        _katex_available = quizcomp.katex.is_available()

    if (inline):
        text = text.strip()

    if (_katex_available):
        content = quizcomp.katex.to_html(text)
    else:
        text = html.escape(text)
        content = f"<code>{text}</code>"

    element = 'span'
    attributes = 'style="margin-left: 0.25em; margin-right: 0.25em"'
    if (not inline):
        element = 'p'
        attributes = 'style="margin-top: 0"'

    return f"<{element} {attributes}>{content}</{element}>"

def _render_md(text, inline, context):
    if (inline):
        text = text.strip()
        return f"$ {text} $"

    return f"$$\n{text}\n$$"



import os
import sys

import quizcomp.args
import quizcomp.converter.convert
import quizcomp.constants
import quizcomp.question.base

def run(args):
    if (not os.path.exists(args.path)):
        raise ValueError(f"Provided path '{args.path}' does not exist.")

    if (not os.path.isfile(args.path)):
        raise ValueError(f"Provided path '{args.path}' is not a file.")

    question = quizcomp.question.base.Question.from_path(args.path)
    content = quizcomp.converter.convert.convert_question(question, format = args.format,
            constructor_args = {'answer_key': args.answer_key})

    print(content)

    return 0

def _get_parser():
    parser = quizcomp.args.Parser(description =
        "Parse a single quiz question (JSON file) and output the result in the specified format.")

    parser.add_argument('path', metavar = 'PATH',
        type = str,
        help = 'The path to a quiz question json file.')

    parser.add_argument('--format',
        action = 'store', type = str, default = quizcomp.constants.FORMAT_JSON,
        choices = quizcomp.converter.convert.SUPPORTED_FORMATS,
        help = 'Output the parsed document in this format (default: %(default)s).')

    parser.add_argument('--key', dest = 'answer_key',
        action = 'store_true', default = False,
        help = 'Generate an answer key instead of a blank quiz (default: %(default)s).')

    return parser

def main():
    args = _get_parser().parse_args()
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())
